"""Test external symbol aliases (EQU external+offset)."""

import pytest
import tempfile
import os


def assemble(source, tmpdir, filename="test.mac"):
    """Assemble source code and return REL bytes."""
    from um80.um80 import Assembler
    src_path = os.path.join(tmpdir, filename)
    with open(src_path, "w") as f:
        f.write(source)
    asm = Assembler()
    if not asm.assemble(src_path):
        raise AssertionError(f"Assembly failed: {asm.errors}")
    return asm.output.get_bytes()


def link(modules):
    """Link multiple modules (list of (name, rel_bytes)) and return output."""
    from um80.ul80 import Linker
    linker = Linker()
    for name, rel_bytes in modules:
        linker.load_rel_data(name, rel_bytes)
    if not linker.link():
        raise AssertionError(f"Link failed: {linker.errors}")
    return linker.output, linker.globals


def test_basic_external_alias():
    """Test basic EQU external+offset alias."""
    with tempfile.TemporaryDirectory() as tmpdir:
        base_src = """\
    .Z80
    CSEG
    PUBLIC  FUNC
FUNC:
    LD  A,10
    ADD A,B
    RET
    END
"""
        alias_src = """\
    .Z80
    CSEG
    EXTRN   FUNC
    PUBLIC  FUNC_ALT
FUNC_ALT    EQU FUNC+2
    END
"""
        main_src = """\
    .Z80
    CSEG
    EXTRN   FUNC
    EXTRN   FUNC_ALT
START:
    CALL    FUNC
    CALL    FUNC_ALT
    RET
    END START
"""
        base_rel = assemble(base_src, tmpdir, "base.mac")
        alias_rel = assemble(alias_src, tmpdir, "alias.mac")
        main_rel = assemble(main_src, tmpdir, "main.mac")

        output, globals_table = link([
            ('MAIN', main_rel),
            ('ALIAS', alias_rel),
            ('BASE', base_rel),
        ])

        # Check that FUNC_ALT is defined and equals FUNC+2
        assert 'FUNC' in globals_table
        assert 'FUNC_ALT' in globals_table

        func_addr = globals_table['FUNC'][1]  # value
        func_alt_addr = globals_table['FUNC_ALT'][1]

        # FUNC_ALT should be FUNC+2
        assert func_alt_addr == func_addr + 2, f"FUNC_ALT ({func_alt_addr}) should be FUNC ({func_addr}) + 2"


def test_external_alias_zero_offset():
    """Test EQU external+0 (simple alias)."""
    with tempfile.TemporaryDirectory() as tmpdir:
        base_src = """\
    .Z80
    CSEG
    PUBLIC  ORIG
ORIG:
    NOP
    RET
    END
"""
        alias_src = """\
    .Z80
    CSEG
    EXTRN   ORIG
    PUBLIC  ALIAS
ALIAS   EQU ORIG
    END
"""
        main_src = """\
    .Z80
    CSEG
    EXTRN   ALIAS
START:
    CALL    ALIAS
    RET
    END START
"""
        base_rel = assemble(base_src, tmpdir, "base.mac")
        alias_rel = assemble(alias_src, tmpdir, "alias.mac")
        main_rel = assemble(main_src, tmpdir, "main.mac")

        output, globals_table = link([
            ('MAIN', main_rel),
            ('ALIAS', alias_rel),
            ('BASE', base_rel),
        ])

        # ALIAS should equal ORIG
        orig_addr = globals_table['ORIG'][1]
        alias_addr = globals_table['ALIAS'][1]
        assert alias_addr == orig_addr


def test_chained_external_alias():
    """Test using an aliased symbol in code that references it with an offset."""
    with tempfile.TemporaryDirectory() as tmpdir:
        base_src = """\
    .Z80
    CSEG
    PUBLIC  BASE
BASE:
    LD  A,1
    LD  A,2
    LD  A,3
    RET
    END
"""
        alias_src = """\
    .Z80
    CSEG
    EXTRN   BASE
    PUBLIC  MID
MID EQU BASE+2
    END
"""
        main_src = """\
    .Z80
    CSEG
    EXTRN   MID
START:
    ; Use MID+2 (which is BASE+4)
    CALL    MID+2
    RET
    END START
"""
        base_rel = assemble(base_src, tmpdir, "base.mac")
        alias_rel = assemble(alias_src, tmpdir, "alias.mac")
        main_rel = assemble(main_src, tmpdir, "main.mac")

        output, globals_table = link([
            ('MAIN', main_rel),
            ('ALIAS', alias_rel),
            ('BASE', base_rel),
        ])

        # MID should be BASE+2
        base_addr = globals_table['BASE'][1]
        mid_addr = globals_table['MID'][1]
        assert mid_addr == base_addr + 2


def test_external_alias_undefined_base():
    """Test that aliased symbol with undefined base produces error."""
    with tempfile.TemporaryDirectory() as tmpdir:
        alias_src = """\
    .Z80
    CSEG
    EXTRN   NONEXISTENT
    PUBLIC  ALIAS
ALIAS   EQU NONEXISTENT+5
    END
"""
        main_src = """\
    .Z80
    CSEG
    EXTRN   ALIAS
START:
    CALL    ALIAS
    RET
    END START
"""
        alias_rel = assemble(alias_src, tmpdir, "alias.mac")
        main_rel = assemble(main_src, tmpdir, "main.mac")

        from um80.ul80 import Linker
        linker = Linker()
        linker.load_rel_data('MAIN', main_rel)
        linker.load_rel_data('ALIAS', alias_rel)

        # Should fail to link
        result = linker.link()
        assert not result
        assert any('NONEXISTENT' in e for e in linker.errors)


if __name__ == '__main__':
    pytest.main([__file__, '-v'])



"""
8080 CPU instruction definitions for um80 assembler.
"""

# Register encodings (3-bit)
REGS = {
    'B': 0, 'C': 1, 'D': 2, 'E': 3, 'H': 4, 'L': 5, 'M': 6, 'A': 7
}

# Register pair encodings (2-bit) for PUSH/POP
REGPAIRS_PUSHPOP = {
    'B': 0, 'BC': 0, 'D': 1, 'DE': 1, 'H': 2, 'HL': 2, 'PSW': 3
}

# Register pair encodings (2-bit) for LXI/DAD/INX/DCX/LDAX/STAX
REGPAIRS = {
    'B': 0, 'BC': 0, 'D': 1, 'DE': 1, 'H': 2, 'HL': 2, 'SP': 3
}

# Register pair for LDAX/STAX (only B and D valid)
REGPAIRS_LDAX = {
    'B': 0, 'BC': 0, 'D': 1, 'DE': 1
}

# Condition codes for jumps/calls/returns
CONDITIONS = {
    'NZ': 0, 'Z': 1, 'NC': 2, 'C': 3, 'PO': 4, 'PE': 5, 'P': 6, 'M': 7
}

# Instructions with no operands
NO_OPERAND = {
    'NOP':  0x00,
    'RLC':  0x07,
    'RRC':  0x0F,
    'RAL':  0x17,
    'RAR':  0x1F,
    'DAA':  0x27,
    'CMA':  0x2F,
    'STC':  0x37,
    'CMC':  0x3F,
    'HLT':  0x76,
    'RET':  0xC9,
    'PCHL': 0xE9,
    'SPHL': 0xF9,
    'XCHG': 0xEB,
    'XTHL': 0xE3,
    'DI':   0xF3,
    'EI':   0xFB,
}

# Conditional returns: RNZ, RZ, RNC, RC, RPO, RPE, RP, RM
# Base opcode is 0xC0, condition in bits 3-5
COND_RET_BASE = 0xC0

# Conditional jumps: JNZ, JZ, JNC, JC, JPO, JPE, JP, JM
# Base opcode is 0xC2, condition in bits 3-5
COND_JMP_BASE = 0xC2

# Conditional calls: CNZ, CZ, CNC, CC, CPO, CPE, CP, CM
# Base opcode is 0xC4, condition in bits 3-5
COND_CALL_BASE = 0xC4

# Unconditional JMP and CALL
JMP = 0xC3
CALL = 0xCD

# RST instruction (RST 0-7)
RST_BASE = 0xC7

# Single register operations
# INR r: 00rrr100, DCR r: 00rrr101
INR_BASE = 0x04
DCR_BASE = 0x05

# Register pair operations
# LXI rp, nn: 00rp0001
# DAD rp: 00rp1001
# INX rp: 00rp0011
# DCX rp: 00rp1011
LXI_BASE = 0x01
DAD_BASE = 0x09
INX_BASE = 0x03
DCX_BASE = 0x0B

# LDAX/STAX (B or D pair only)
# LDAX rp: 00rp1010
# STAX rp: 00rp0010
LDAX_BASE = 0x0A
STAX_BASE = 0x02

# PUSH/POP (B, D, H, PSW)
# PUSH rp: 11rp0101
# POP rp: 11rp0001
PUSH_BASE = 0xC5
POP_BASE = 0xC1

# MOV dst, src: 01dddsss
MOV_BASE = 0x40

# MVI r, n: 00rrr110
MVI_BASE = 0x06

# Arithmetic/logical with register
# ADD r: 10000rrr, ADC r: 10001rrr, SUB r: 10010rrr, SBB r: 10011rrr
# ANA r: 10100rrr, XRA r: 10101rrr, ORA r: 10110rrr, CMP r: 10111rrr
ALU_REG = {
    'ADD': 0x80, 'ADC': 0x88, 'SUB': 0x90, 'SBB': 0x98,
    'ANA': 0xA0, 'XRA': 0xA8, 'ORA': 0xB0, 'CMP': 0xB8
}

# Arithmetic/logical immediate
# ADI n: 0xC6, ACI n: 0xCE, SUI n: 0xD6, SBI n: 0xDE
# ANI n: 0xE6, XRI n: 0xEE, ORI n: 0xF6, CPI n: 0xFE
ALU_IMM = {
    'ADI': 0xC6, 'ACI': 0xCE, 'SUI': 0xD6, 'SBI': 0xDE,
    'ANI': 0xE6, 'XRI': 0xEE, 'ORI': 0xF6, 'CPI': 0xFE
}

# Memory reference with 16-bit address
# LDA addr: 0x3A, STA addr: 0x32
# LHLD addr: 0x2A, SHLD addr: 0x22
LDA = 0x3A
STA = 0x32
LHLD = 0x2A
SHLD = 0x22

# I/O instructions
# IN port: 0xDB, OUT port: 0xD3
IN = 0xDB
OUT = 0xD3


def encode_no_operand(mnemonic):
    """Encode instruction with no operands."""
    return bytes([NO_OPERAND[mnemonic.upper()]])


def encode_mov(dst, src):
    """Encode MOV dst, src instruction."""
    dst_code = REGS[dst.upper()]
    src_code = REGS[src.upper()]
    return bytes([MOV_BASE | (dst_code << 3) | src_code])


def encode_mvi(reg, value):
    """Encode MVI reg, value instruction."""
    reg_code = REGS[reg.upper()]
    return bytes([MVI_BASE | (reg_code << 3), value & 0xFF])


def encode_lxi(regpair, value):
    """Encode LXI regpair, value instruction."""
    rp_code = REGPAIRS[regpair.upper()]
    return bytes([LXI_BASE | (rp_code << 4), value & 0xFF, (value >> 8) & 0xFF])


def encode_inr(reg):
    """Encode INR reg instruction."""
    reg_code = REGS[reg.upper()]
    return bytes([INR_BASE | (reg_code << 3)])


def encode_dcr(reg):
    """Encode DCR reg instruction."""
    reg_code = REGS[reg.upper()]
    return bytes([DCR_BASE | (reg_code << 3)])


def encode_inx(regpair):
    """Encode INX regpair instruction."""
    rp_code = REGPAIRS[regpair.upper()]
    return bytes([INX_BASE | (rp_code << 4)])


def encode_dcx(regpair):
    """Encode DCX regpair instruction."""
    rp_code = REGPAIRS[regpair.upper()]
    return bytes([DCX_BASE | (rp_code << 4)])


def encode_dad(regpair):
    """Encode DAD regpair instruction."""
    rp_code = REGPAIRS[regpair.upper()]
    return bytes([DAD_BASE | (rp_code << 4)])


def encode_ldax(regpair):
    """Encode LDAX regpair instruction (B or D only)."""
    rp_code = REGPAIRS_LDAX[regpair.upper()]
    return bytes([LDAX_BASE | (rp_code << 4)])


def encode_stax(regpair):
    """Encode STAX regpair instruction (B or D only)."""
    rp_code = REGPAIRS_LDAX[regpair.upper()]
    return bytes([STAX_BASE | (rp_code << 4)])


def encode_push(regpair):
    """Encode PUSH regpair instruction."""
    rp_code = REGPAIRS_PUSHPOP[regpair.upper()]
    return bytes([PUSH_BASE | (rp_code << 4)])


def encode_pop(regpair):
    """Encode POP regpair instruction."""
    rp_code = REGPAIRS_PUSHPOP[regpair.upper()]
    return bytes([POP_BASE | (rp_code << 4)])


def encode_alu_reg(mnemonic, reg):
    """Encode ALU operation with register (ADD, ADC, SUB, etc.)."""
    base = ALU_REG[mnemonic.upper()]
    reg_code = REGS[reg.upper()]
    return bytes([base | reg_code])


def encode_alu_imm(mnemonic, value):
    """Encode ALU immediate operation (ADI, ACI, SUI, etc.)."""
    opcode = ALU_IMM[mnemonic.upper()]
    return bytes([opcode, value & 0xFF])


def encode_jmp(addr):
    """Encode unconditional JMP."""
    return bytes([JMP, addr & 0xFF, (addr >> 8) & 0xFF])


def encode_cond_jmp(cond, addr):
    """Encode conditional jump (JNZ, JZ, etc.)."""
    cond_code = CONDITIONS[cond.upper()]
    opcode = COND_JMP_BASE | (cond_code << 3)
    return bytes([opcode, addr & 0xFF, (addr >> 8) & 0xFF])


def encode_call(addr):
    """Encode unconditional CALL."""
    return bytes([CALL, addr & 0xFF, (addr >> 8) & 0xFF])


def encode_cond_call(cond, addr):
    """Encode conditional call (CNZ, CZ, etc.)."""
    cond_code = CONDITIONS[cond.upper()]
    opcode = COND_CALL_BASE | (cond_code << 3)
    return bytes([opcode, addr & 0xFF, (addr >> 8) & 0xFF])


def encode_cond_ret(cond):
    """Encode conditional return (RNZ, RZ, etc.)."""
    cond_code = CONDITIONS[cond.upper()]
    return bytes([COND_RET_BASE | (cond_code << 3)])


def encode_rst(n):
    """Encode RST n instruction (n = 0-7)."""
    return bytes([RST_BASE | ((n & 7) << 3)])


def encode_lda(addr):
    """Encode LDA addr."""
    return bytes([LDA, addr & 0xFF, (addr >> 8) & 0xFF])


def encode_sta(addr):
    """Encode STA addr."""
    return bytes([STA, addr & 0xFF, (addr >> 8) & 0xFF])


def encode_lhld(addr):
    """Encode LHLD addr."""
    return bytes([LHLD, addr & 0xFF, (addr >> 8) & 0xFF])


def encode_shld(addr):
    """Encode SHLD addr."""
    return bytes([SHLD, addr & 0xFF, (addr >> 8) & 0xFF])


def encode_in(port):
    """Encode IN port."""
    return bytes([IN, port & 0xFF])


def encode_out(port):
    """Encode OUT port."""
    return bytes([OUT, port & 0xFF])


# Conditional jump mnemonics
COND_JUMPS = {'JNZ', 'JZ', 'JNC', 'JC', 'JPO', 'JPE', 'JP', 'JM'}

# Conditional call mnemonics
COND_CALLS = {'CNZ', 'CZ', 'CNC', 'CC', 'CPO', 'CPE', 'CP', 'CM'}

# Conditional return mnemonics
COND_RETS = {'RNZ', 'RZ', 'RNC', 'RC', 'RPO', 'RPE', 'RP', 'RM'}

# Opcodes as operand values (first byte of instruction encoding)
# This allows using opcode names as values in expressions per M80 manual p.2-4
OPCODE_VALUES = {
    # No-operand instructions
    'NOP': 0x00, 'RLC': 0x07, 'RRC': 0x0F, 'RAL': 0x17, 'RAR': 0x1F,
    'DAA': 0x27, 'CMA': 0x2F, 'STC': 0x37, 'CMC': 0x3F, 'HLT': 0x76,
    'RET': 0xC9, 'PCHL': 0xE9, 'SPHL': 0xF9, 'XCHG': 0xEB, 'XTHL': 0xE3,
    'DI': 0xF3, 'EI': 0xFB,
    # Conditional returns
    'RNZ': 0xC0, 'RZ': 0xC8, 'RNC': 0xD0, 'RC': 0xD8,
    'RPO': 0xE0, 'RPE': 0xE8, 'RP': 0xF0, 'RM': 0xF8,
    # Conditional jumps
    'JNZ': 0xC2, 'JZ': 0xCA, 'JNC': 0xD2, 'JC': 0xDA,
    'JPO': 0xE2, 'JPE': 0xEA, 'JP': 0xF2, 'JM': 0xFA,
    # Unconditional jump/call
    'JMP': 0xC3, 'CALL': 0xCD,
    # Conditional calls
    'CNZ': 0xC4, 'CZ': 0xCC, 'CNC': 0xD4, 'CC': 0xDC,
    'CPO': 0xE4, 'CPE': 0xEC, 'CP': 0xF4, 'CM': 0xFC,
    # ALU immediate
    'ADI': 0xC6, 'ACI': 0xCE, 'SUI': 0xD6, 'SBI': 0xDE,
    'ANI': 0xE6, 'XRI': 0xEE, 'ORI': 0xF6, 'CPI': 0xFE,
    # Memory reference
    'LDA': 0x3A, 'STA': 0x32, 'LHLD': 0x2A, 'SHLD': 0x22,
    # I/O
    'IN': 0xDB, 'OUT': 0xD3,
    # RST base (RST 0)
    'RST': 0xC7,
    # Register operations bases (use with register encoding)
    'MOV': 0x40, 'MVI': 0x06,
    'INR': 0x04, 'DCR': 0x05,
    'ADD': 0x80, 'ADC': 0x88, 'SUB': 0x90, 'SBB': 0x98,
    'ANA': 0xA0, 'XRA': 0xA8, 'ORA': 0xB0, 'CMP': 0xB8,
    # Register pair operations bases
    'LXI': 0x01, 'DAD': 0x09, 'INX': 0x03, 'DCX': 0x0B,
    'PUSH': 0xC5, 'POP': 0xC1,
    'LDAX': 0x0A, 'STAX': 0x02,
}


def get_cond_from_mnemonic(mnemonic):
    """Extract condition code from mnemonic like JNZ -> NZ."""
    m = mnemonic.upper()
    if m.startswith('J'):
        return m[1:]
    elif m.startswith('C') and m != 'CALL' and m != 'CMP' and m != 'CMA' and m != 'CMC':
        return m[1:]
    elif m.startswith('R') and m != 'RET' and m != 'RLC' and m != 'RRC' and m != 'RAL' and m != 'RAR' and m != 'RST':
        return m[1:]
    return None



"""
Test case sensitivity handling in the um80 toolchain.

The original Microsoft L80 linker was case-insensitive - all symbols were
converted to uppercase internally. This test verifies that behavior.
"""

import os
import tempfile
import pytest
from um80.relformat import RELWriter, RELReader


class TestRELFormatCaseSensitivity:
    """Test that REL format handles symbol case correctly."""

    def test_read_b_field_uppercases_symbols(self):
        """Verify _read_b_field returns uppercase symbols for L80 compatibility."""
        # Write a REL file with a lowercase symbol using the writer
        # (which uppercases on write), then verify reader also uppercases
        writer = RELWriter()
        writer.write_entry_symbol("lowercase")
        writer.write_end_file()
        data = writer.get_bytes()

        reader = RELReader(data)
        item = reader.read_item()

        assert item[0] == 'ENTRY_SYMBOL'
        assert item[1] == 'LOWERCASE'  # Uppercased (no truncation by default)

    def test_mixed_case_symbols_normalized(self):
        """Verify mixed case symbols are normalized to uppercase."""
        writer = RELWriter()
        writer.write_entry_symbol("MixedCase")
        writer.write_end_file()
        data = writer.get_bytes()

        reader = RELReader(data)
        item = reader.read_item()

        assert item[0] == 'ENTRY_SYMBOL'
        assert item[1] == 'MIXEDCASE'

    def test_already_uppercase_unchanged(self):
        """Verify uppercase symbols pass through correctly."""
        writer = RELWriter()
        writer.write_entry_symbol("ALREADY")
        writer.write_end_file()
        data = writer.get_bytes()

        reader = RELReader(data)
        item = reader.read_item()

        assert item[0] == 'ENTRY_SYMBOL'
        assert item[1] == 'ALREADY'

    def test_external_chain_uppercase(self):
        """Verify CHAIN_EXTERNAL symbol names are uppercased."""
        writer = RELWriter()
        writer.write_chain_external(0, 0x1234, "extfunc")
        writer.write_end_file()
        data = writer.get_bytes()

        reader = RELReader(data)
        item = reader.read_item()

        assert item[0] == 'CHAIN_EXTERNAL'
        assert item[2] == 'EXTFUNC'

    def test_define_entry_uppercase(self):
        """Verify DEFINE_ENTRY (PUBLIC) symbol names are uppercased."""
        writer = RELWriter()
        writer.write_define_entry_point(1, 0x100, "myPublic")
        writer.write_end_file()
        data = writer.get_bytes()

        reader = RELReader(data)
        item = reader.read_item()

        assert item[0] == 'DEFINE_ENTRY'
        assert item[2] == 'MYPUBLIC'

    def test_common_block_uppercase(self):
        """Verify COMMON block names are uppercased."""
        writer = RELWriter()
        writer.write_define_common_size(0, 100, "myCommon")
        writer.write_end_file()
        data = writer.get_bytes()

        reader = RELReader(data)
        item = reader.read_item()

        assert item[0] == 'DEFINE_COMMON_SIZE'
        assert item[2] == 'MYCOMMON'

    def test_program_name_uppercase(self):
        """Verify program names are uppercased."""
        writer = RELWriter()
        writer.write_program_name("myModule")
        writer.write_end_file()
        data = writer.get_bytes()

        reader = RELReader(data)
        item = reader.read_item()

        assert item[0] == 'PROGRAM_NAME'
        assert item[1] == 'MYMODULE'


class TestLinkerCaseSensitivity:
    """Integration tests for linker case sensitivity."""

    def test_linker_resolves_mixed_case_externals(self):
        """Verify linker resolves external references case-insensitively."""
        from um80.ul80 import Linker

        with tempfile.TemporaryDirectory() as tmpdir:
            # Create module1.rel with lowercase public symbol
            writer1 = RELWriter()
            writer1.write_program_name("MOD1")
            writer1.write_define_entry_point(1, 0, "mysymbol")  # lowercase
            writer1.write_define_program_size(3)
            writer1.write_absolute_byte(0xC9)  # RET
            writer1.write_absolute_byte(0x00)
            writer1.write_absolute_byte(0x00)
            writer1.write_end_program()
            writer1.write_end_file()

            mod1_path = os.path.join(tmpdir, "mod1.rel")
            with open(mod1_path, "wb") as f:
                f.write(writer1.get_bytes())

            # Create module2.rel with uppercase external reference
            writer2 = RELWriter()
            writer2.write_program_name("MOD2")
            writer2.write_chain_external(0, 1, "MYSYMBOL")  # uppercase
            writer2.write_define_program_size(4)
            writer2.write_absolute_byte(0xCD)  # CALL
            writer2.write_absolute_byte(0x00)  # placeholder
            writer2.write_absolute_byte(0x00)
            writer2.write_absolute_byte(0xC9)  # RET
            writer2.write_end_program()
            writer2.write_end_file()

            mod2_path = os.path.join(tmpdir, "mod2.rel")
            with open(mod2_path, "wb") as f:
                f.write(writer2.get_bytes())

            # Link them
            linker = Linker()
            linker.load_rel(mod1_path)
            linker.load_rel(mod2_path)

            # Should resolve without "undefined symbol" error
            assert linker.resolve_externals() is True

    def test_sym_file_contains_uppercase(self):
        """Verify .SYM output contains uppercase symbols."""
        from um80.ul80 import Linker

        with tempfile.TemporaryDirectory() as tmpdir:
            # Create a module with mixed case symbols
            writer = RELWriter()
            writer.write_program_name("TEST")
            writer.write_define_entry_point(1, 0, "LowerCase")
            writer.write_define_entry_point(1, 3, "UPPERCASE")
            writer.write_define_entry_point(1, 6, "mixedCase")
            writer.write_define_program_size(9)
            for _ in range(9):
                writer.write_absolute_byte(0xC9)
            writer.write_end_program()
            writer.write_end_file()

            rel_path = os.path.join(tmpdir, "test.rel")
            with open(rel_path, "wb") as f:
                f.write(writer.get_bytes())

            # Link and generate symbol file
            linker = Linker()
            linker.load_rel(rel_path)
            linker.resolve_externals()
            linker.calculate_addresses()

            sym_path = os.path.join(tmpdir, "test.sym")
            linker.save_sym(sym_path)

            # Read and verify symbol file
            with open(sym_path, "r") as f:
                content = f.read()

            # All symbols should be uppercase
            assert "LOWERCAS" in content
            assert "UPPERCAS" in content
            assert "MIXEDCAS" in content
            # No lowercase versions should exist
            assert "LowerCase" not in content
            assert "mixedCase" not in content


if __name__ == "__main__":
    pytest.main([__file__, "-v"])



"""
um80 - Microsoft MACRO-80 compatible assembler toolchain for Linux.

This package provides Unix/Linux implementations of the classic CP/M
development tools from Microsoft:

- um80: MACRO-80 compatible assembler
- ul80: LINK-80 compatible linker
- ulib80: LIB-80 compatible library manager
- ucref80: Cross-reference utility
- ud80: 8080/Z80 disassembler
- ux80: 8080 to Z80 assembly translator
"""

from importlib.metadata import version, PackageNotFoundError

try:
    __version__ = version("um80")
except PackageNotFoundError:
    __version__ = "unknown"



#!/usr/bin/env python3
"""
ul80 - Microsoft LINK-80 compatible linker for Linux.

Usage: ul80 [-o output.com] file1.rel file2.rel ... [lib1.lib ...]

Supports both .rel object files and .lib library files (created by ulib80).
Library modules are automatically extracted to resolve undefined symbols.
"""

import sys
import os
import argparse
from pathlib import Path

from um80 import __version__
from um80.relformat import *
from um80.ulib80 import Library, LibraryError


class LinkerError(Exception):
    """Linker error."""
    pass


class Module:
    """A loaded REL module."""
    def __init__(self, name):
        self.name = name
        self.code = bytearray()  # Loaded code/data
        self.code_size = 0
        self.data_size = 0
        self.code_base = 0  # Will be set during linking
        self.data_base = 0
        self.code_start = 0  # Starting offset within code buffer (for absolute ORG)

        # Symbols defined in this module
        self.publics = {}  # name -> (value, seg_type)

        # Aliased entry points: SYMBOL EQU EXTERNAL+offset made PUBLIC
        # These are resolved after all externals are resolved
        self.aliased_publics = {}  # new_name -> (base_external, offset)

        # External references (chains to be fixed up)
        self.externals = {}  # name -> list of (offset_in_module, seg_type)

        # Internal chains (forward references resolved within module)
        self.chains = {}  # offset -> list of offsets to fix

        # Common blocks
        self.commons = {}  # name -> size

        # Relocation info: list of (offset, seg_type) for addresses that need relocation
        self.relocations = []

        # Segment buffer offsets: maps segment type to buffer offset where that segment starts
        # Used to convert segment-relative addresses to buffer offsets during chain following
        self.seg_buf_start = {}  # seg_type -> buffer start offset


class Linker:
    """LINK-80 compatible linker."""

    def __init__(self):
        self.modules = []
        self.globals = {}  # name -> (module_idx, value, seg_type, is_defined)
        self.commons = {}  # name -> size (largest wins)

        # Pre-define linker symbols (values computed in calculate_addresses)
        # mod_idx=0 is placeholder, value will be absolute address
        self.globals['__END__'] = (0, 0, ADDR_ABSOLUTE, True)
        self.globals['__BSS_START'] = (0, 0, ADDR_ABSOLUTE, True)
        self.globals['__BSS_END'] = (0, 0, ADDR_ABSOLUTE, True)

        self.code_base = 0x0103  # Default CP/M load address + 3 for JMP
        self.data_base = None  # Will be after code if not specified
        self.common_base = None  # After data

        self.output = bytearray()
        self.entry_point = None  # (value, seg_type) or None

        self.errors = []

        # External relocations: output offsets (low byte position) that need
        # relocation due to external symbol resolution to CSEG symbols.
        # Populated by resolve_externals(), used by save_prl().
        self.external_relocations = []

        # When True, emit zeros for DS (reserve space) directives instead of
        # treating them as BSS. Required for PRL/SPR format where all segments
        # must be contiguous in the output. Default is True for compatibility.
        self.emit_ds_zeros = True
        self.warnings = []

    def error(self, msg):
        self.errors.append(f"Error: {msg}")

    def warning(self, msg):
        self.warnings.append(f"Warning: {msg}")

    def load_rel(self, filename):
        """Load a REL file and add to modules list."""
        with open(filename, 'rb') as f:
            data = f.read()

        reader = RELReader(data)
        module = Module(Path(filename).stem.upper())

        current_loc = 0  # Position within current segment
        current_seg = ADDR_PROGRAM_REL  # Default to code segment
        first_loc_set = False  # Track if we've seen the first location

        # Use separate buffers for each segment to avoid overwrites when
        # switching between segments (e.g., CSEG -> DSEG -> CSEG)
        seg_buffers = {}  # seg_type -> bytearray

        def get_seg_buffer():
            """Get or create buffer for current segment."""
            if current_seg not in seg_buffers:
                seg_buffers[current_seg] = bytearray()
            return seg_buffers[current_seg]

        def write_byte_to_seg(value):
            """Write a byte at current_loc in current segment's buffer."""
            buf = get_seg_buffer()
            while len(buf) <= current_loc:
                buf.append(0)
            buf[current_loc] = value

        # Track relocations with segment-relative offsets before combining
        pending_relocations = []  # (seg_type, seg_offset, reloc_type)
        pending_externals = []  # (name, seg_type, head_offset)
        pending_chains = []  # (chain_seg, head_offset, cur_seg, cur_offset)

        while True:
            try:
                item = reader.read_item()
            except EOFError:
                break

            if item is None:
                break

            item_type = item[0]

            if item_type == 'ABSOLUTE_BYTE':
                write_byte_to_seg(item[1])
                current_loc += 1

            elif item_type == 'PROGRAM_REL':
                # 16-bit program-relative value - needs relocation
                value = item[1]
                write_byte_to_seg(value & 0xFF)
                # Record relocation at low byte position
                pending_relocations.append((current_seg, current_loc, ADDR_PROGRAM_REL))
                current_loc += 1
                write_byte_to_seg((value >> 8) & 0xFF)
                current_loc += 1

            elif item_type == 'DATA_REL':
                # 16-bit data-relative value - needs relocation
                value = item[1]
                write_byte_to_seg(value & 0xFF)
                pending_relocations.append((current_seg, current_loc, ADDR_DATA_REL))
                current_loc += 1
                write_byte_to_seg((value >> 8) & 0xFF)
                current_loc += 1

            elif item_type == 'COMMON_REL':
                # 16-bit common-relative value - needs relocation
                value = item[1]
                write_byte_to_seg(value & 0xFF)
                pending_relocations.append((current_seg, current_loc, ADDR_COMMON_REL))
                current_loc += 1
                write_byte_to_seg((value >> 8) & 0xFF)
                current_loc += 1

            elif item_type == 'PROGRAM_NAME':
                module.name = item[1]

            elif item_type == 'ENTRY_SYMBOL':
                # Symbol this module exports (for library search)
                pass

            elif item_type == 'DEFINE_ENTRY':
                # PUBLIC symbol definition
                a_field, name = item[1], item[2]
                addr_type, value = a_field
                # Check for aliased entry (format: "NEWNAME=EXTERNAL" or "NEWNAME=EXTERNAL+N")
                if '=' in name:
                    new_name, alias_spec = name.split('=', 1)
                    # Parse the alias spec: "EXTERNAL" or "EXTERNAL+N"
                    if '+' in alias_spec:
                        base_ext, offset_str = alias_spec.rsplit('+', 1)
                        try:
                            offset = int(offset_str)
                        except ValueError:
                            offset = 0
                            base_ext = alias_spec
                    else:
                        base_ext = alias_spec
                        offset = 0
                    module.aliased_publics[new_name] = (base_ext, offset)
                else:
                    module.publics[name] = (value, addr_type)

            elif item_type == 'CHAIN_EXTERNAL':
                # External reference chain - store segment-relative for now
                a_field, name = item[1], item[2]
                addr_type, head = a_field
                pending_externals.append((name, addr_type, head))

            elif item_type == 'SET_LOC':
                a_field = item[1]
                addr_type, value = a_field
                # If emit_ds_zeros is enabled, fill gaps with zeros
                # This handles DS directives which advance without emitting bytes
                if self.emit_ds_zeros and value > 0:
                    # Switch to target segment first to write to correct buffer
                    current_seg = addr_type
                    # Get current buffer size for this segment
                    buf = get_seg_buffer()
                    fill_from = len(buf)
                    if value > fill_from:
                        # Fill zeros from current buffer end to new position
                        current_loc = fill_from
                        while current_loc < value:
                            write_byte_to_seg(0)
                            current_loc += 1
                current_loc = value
                current_seg = addr_type
                # Track the first absolute location as code_start
                if not first_loc_set and addr_type == ADDR_ABSOLUTE:
                    module.code_start = value
                    first_loc_set = True

            elif item_type == 'CHAIN_ADDRESS':
                # Internal forward reference chain - store segment-relative
                a_field = item[1]
                addr_type, head = a_field
                pending_chains.append((addr_type, head, current_seg, current_loc))

            elif item_type == 'DEFINE_PROG_SIZE':
                a_field = item[1]
                _, size = a_field
                module.code_size = size

            elif item_type == 'DEFINE_DATA_SIZE':
                a_field = item[1]
                _, size = a_field
                module.data_size = size

            elif item_type == 'DEFINE_COMMON_SIZE':
                a_field, name = item[1], item[2]
                _, size = a_field
                module.commons[name] = size

            elif item_type == 'SELECT_COMMON':
                # Switch to common block
                pass

            elif item_type == 'REQUEST_LIB':
                # Library search request
                pass

            elif item_type == 'END_PROGRAM':
                # End of module
                pass

            elif item_type == 'END_FILE':
                break

        # Combine segment buffers into single code buffer
        # Order: ASEG (absolute), CSEG (program), DSEG (data), COMMON
        code_bytes = bytearray()
        seg_buf_start = {}

        for seg_type in [ADDR_ABSOLUTE, ADDR_PROGRAM_REL, ADDR_DATA_REL, ADDR_COMMON_REL]:
            if seg_type in seg_buffers:
                seg_buf_start[seg_type] = len(code_bytes)
                code_bytes.extend(seg_buffers[seg_type])

        # Convert segment-relative relocations to buffer offsets
        # Store (buf_offset, reloc_type, ref_seg_type) where ref_seg_type is which segment the reference is in
        for seg_type, seg_offset, reloc_type in pending_relocations:
            if seg_type in seg_buf_start:
                buf_offset = seg_buf_start[seg_type] + seg_offset
                module.relocations.append((buf_offset, reloc_type, seg_type))

        # Convert pending externals to buffer offsets
        for name, addr_type, head in pending_externals:
            if name not in module.externals:
                module.externals[name] = []
            if addr_type in seg_buf_start:
                buf_head = seg_buf_start[addr_type] + head
            else:
                buf_head = seg_buf_start.get(ADDR_ABSOLUTE, len(code_bytes)) + head
            module.externals[name].append((buf_head, addr_type))

        # Convert pending chains to buffer offsets
        for chain_seg, head, cur_seg, cur_offset in pending_chains:
            if chain_seg in seg_buf_start:
                buf_head = seg_buf_start[chain_seg] + head
            else:
                buf_head = len(code_bytes) + head
            if cur_seg in seg_buf_start:
                buf_cur = seg_buf_start[cur_seg] + cur_offset
            else:
                buf_cur = len(code_bytes) + cur_offset
            if buf_head not in module.chains:
                module.chains[buf_head] = []
            module.chains[buf_head].append((buf_cur, chain_seg))

        module.code = code_bytes
        module.seg_buf_start = seg_buf_start  # Save for chain following during link
        self.modules.append(module)

        # Register public symbols
        mod_idx = len(self.modules) - 1
        for name, (value, seg_type) in module.publics.items():
            if name in self.globals and self.globals[name][3]:
                self.warning(f"Multiple definition of '{name}'")
            else:
                self.globals[name] = (mod_idx, value, seg_type, True)

        # Track common block sizes
        for name, size in module.commons.items():
            if name not in self.commons or size > self.commons[name]:
                self.commons[name] = size

        return True

    def load_rel_data(self, name, data):
        """Load REL data from bytes (e.g., from a library module)."""
        reader = RELReader(data)
        module = Module(name.upper())

        current_loc = 0  # Position within current segment
        current_seg = ADDR_PROGRAM_REL  # Default to code segment
        first_loc_set = False  # Track if we've seen the first location

        # Use separate buffers for each segment to avoid overwrites when
        # switching between segments (e.g., CSEG -> DSEG -> CSEG)
        seg_buffers = {}  # seg_type -> bytearray

        def get_seg_buffer():
            """Get or create buffer for current segment."""
            if current_seg not in seg_buffers:
                seg_buffers[current_seg] = bytearray()
            return seg_buffers[current_seg]

        def write_byte_to_seg(value):
            """Write a byte at current_loc in current segment's buffer."""
            buf = get_seg_buffer()
            while len(buf) <= current_loc:
                buf.append(0)
            buf[current_loc] = value

        # Track relocations with segment-relative offsets before combining
        pending_relocations = []  # (seg_type, seg_offset, reloc_type)
        pending_externals = []  # (name, seg_type, head_offset)
        pending_chains = []  # (chain_seg, head_offset, cur_seg, cur_offset)

        while True:
            try:
                item = reader.read_item()
            except EOFError:
                break

            if item is None:
                break

            item_type = item[0]

            if item_type == 'ABSOLUTE_BYTE':
                write_byte_to_seg(item[1])
                current_loc += 1

            elif item_type == 'PROGRAM_REL':
                # 16-bit program-relative value - needs relocation
                value = item[1]
                write_byte_to_seg(value & 0xFF)
                # Record relocation at low byte position
                pending_relocations.append((current_seg, current_loc, ADDR_PROGRAM_REL))
                current_loc += 1
                write_byte_to_seg((value >> 8) & 0xFF)
                current_loc += 1

            elif item_type == 'DATA_REL':
                # 16-bit data-relative value - needs relocation
                value = item[1]
                write_byte_to_seg(value & 0xFF)
                pending_relocations.append((current_seg, current_loc, ADDR_DATA_REL))
                current_loc += 1
                write_byte_to_seg((value >> 8) & 0xFF)
                current_loc += 1

            elif item_type == 'COMMON_REL':
                # 16-bit common-relative value - needs relocation
                value = item[1]
                write_byte_to_seg(value & 0xFF)
                pending_relocations.append((current_seg, current_loc, ADDR_COMMON_REL))
                current_loc += 1
                write_byte_to_seg((value >> 8) & 0xFF)
                current_loc += 1

            elif item_type == 'PROGRAM_NAME':
                module.name = item[1]

            elif item_type == 'ENTRY_SYMBOL':
                # Symbol this module exports (for library search)
                pass

            elif item_type == 'DEFINE_ENTRY':
                # PUBLIC symbol definition
                a_field, sym_name = item[1], item[2]
                addr_type, value = a_field
                # Check for aliased entry (format: "NEWNAME=EXTERNAL" or "NEWNAME=EXTERNAL+N")
                if '=' in sym_name:
                    new_name, alias_spec = sym_name.split('=', 1)
                    # Parse the alias spec: "EXTERNAL" or "EXTERNAL+N"
                    if '+' in alias_spec:
                        base_ext, offset_str = alias_spec.rsplit('+', 1)
                        try:
                            offset = int(offset_str)
                        except ValueError:
                            offset = 0
                            base_ext = alias_spec
                    else:
                        base_ext = alias_spec
                        offset = 0
                    module.aliased_publics[new_name] = (base_ext, offset)
                else:
                    module.publics[sym_name] = (value, addr_type)

            elif item_type == 'CHAIN_EXTERNAL':
                # External reference chain - store segment-relative for now
                a_field, sym_name = item[1], item[2]
                addr_type, head = a_field
                pending_externals.append((sym_name, addr_type, head))

            elif item_type == 'SET_LOC':
                a_field = item[1]
                addr_type, value = a_field
                # If emit_ds_zeros is enabled, fill gaps with zeros
                # This handles DS directives which advance without emitting bytes
                if self.emit_ds_zeros and value > 0:
                    # Switch to target segment first to write to correct buffer
                    current_seg = addr_type
                    # Get current buffer size for this segment
                    buf = get_seg_buffer()
                    fill_from = len(buf)
                    if value > fill_from:
                        # Temporarily set current_loc to fill position
                        current_loc = fill_from
                        while current_loc < value:
                            write_byte_to_seg(0)
                            current_loc += 1
                current_loc = value
                current_seg = addr_type
                # Track the first absolute location as code_start
                if not first_loc_set and addr_type == ADDR_ABSOLUTE:
                    module.code_start = value
                    first_loc_set = True

            elif item_type == 'CHAIN_ADDRESS':
                # Internal forward reference chain - store segment-relative
                a_field = item[1]
                addr_type, head = a_field
                pending_chains.append((addr_type, head, current_seg, current_loc))

            elif item_type == 'DEFINE_PROG_SIZE':
                a_field = item[1]
                _, size = a_field
                module.code_size = size

            elif item_type == 'DEFINE_DATA_SIZE':
                a_field = item[1]
                _, size = a_field
                module.data_size = size

            elif item_type == 'DEFINE_COMMON_SIZE':
                a_field, sym_name = item[1], item[2]
                _, size = a_field
                module.commons[sym_name] = size

            elif item_type == 'SELECT_COMMON':
                # Switch to common block
                pass

            elif item_type == 'REQUEST_LIB':
                # Library search request
                pass

            elif item_type == 'END_PROGRAM':
                # End of module
                pass

            elif item_type == 'END_FILE':
                break

        # Combine segment buffers into single code buffer
        # Order: ASEG (absolute), CSEG (program), DSEG (data), COMMON
        code_bytes = bytearray()
        seg_buf_start = {}

        for seg_type in [ADDR_ABSOLUTE, ADDR_PROGRAM_REL, ADDR_DATA_REL, ADDR_COMMON_REL]:
            if seg_type in seg_buffers:
                seg_buf_start[seg_type] = len(code_bytes)
                code_bytes.extend(seg_buffers[seg_type])

        # Convert segment-relative relocations to buffer offsets
        # Store (buf_offset, reloc_type, ref_seg_type) where ref_seg_type is which segment the reference is in
        for seg_type, seg_offset, reloc_type in pending_relocations:
            if seg_type in seg_buf_start:
                buf_offset = seg_buf_start[seg_type] + seg_offset
                module.relocations.append((buf_offset, reloc_type, seg_type))

        # Convert pending externals to buffer offsets
        for sym_name, addr_type, head in pending_externals:
            if sym_name not in module.externals:
                module.externals[sym_name] = []
            if addr_type in seg_buf_start:
                buf_head = seg_buf_start[addr_type] + head
            else:
                buf_head = seg_buf_start.get(ADDR_ABSOLUTE, len(code_bytes)) + head
            module.externals[sym_name].append((buf_head, addr_type))

        # Convert pending chains to buffer offsets
        for chain_seg, head, cur_seg, cur_offset in pending_chains:
            if chain_seg in seg_buf_start:
                buf_head = seg_buf_start[chain_seg] + head
            else:
                buf_head = len(code_bytes) + head
            if cur_seg in seg_buf_start:
                buf_cur = seg_buf_start[cur_seg] + cur_offset
            else:
                buf_cur = len(code_bytes) + cur_offset
            if buf_head not in module.chains:
                module.chains[buf_head] = []
            module.chains[buf_head].append((buf_cur, chain_seg))

        module.code = code_bytes
        module.seg_buf_start = seg_buf_start  # Save for chain following during link
        self.modules.append(module)

        # Register public symbols
        mod_idx = len(self.modules) - 1
        for sym_name, (value, seg_type) in module.publics.items():
            if sym_name in self.globals and self.globals[sym_name][3]:
                self.warning(f"Multiple definition of '{sym_name}'")
            else:
                self.globals[sym_name] = (mod_idx, value, seg_type, True)

        # Track common block sizes
        for sym_name, size in module.commons.items():
            if sym_name not in self.commons or size > self.commons[sym_name]:
                self.commons[sym_name] = size

        return True

    def get_undefined_symbols(self):
        """Get list of undefined external symbols."""
        undefined = set()
        for module in self.modules:
            for name in module.externals:
                # Parse "SYMBOL+N" format - check base symbol
                base_name = name
                if '+' in name:
                    parts = name.rsplit('+', 1)
                    try:
                        int(parts[1])  # Valid offset?
                        base_name = parts[0]
                    except ValueError:
                        pass  # Not a valid offset, use full name

                if base_name not in self.globals or not self.globals[base_name][3]:
                    undefined.add(base_name)
        return undefined

    def resolve_externals(self):
        """Check that all external references can be resolved."""
        undefined = []
        for module in self.modules:
            for name in module.externals:
                # Parse "SYMBOL+N" format - check base symbol
                base_name = name
                if '+' in name:
                    parts = name.rsplit('+', 1)
                    try:
                        int(parts[1])  # Valid offset?
                        base_name = parts[0]
                    except ValueError:
                        pass  # Not a valid offset, use full name

                if base_name not in self.globals or not self.globals[base_name][3]:
                    undefined.append(name)

        if undefined:
            for name in set(undefined):
                self.error(f"Undefined symbol: {name}")
            return False
        return True

    def resolve_aliased_publics(self):
        """Resolve aliased public symbols (EQU external+offset made PUBLIC).

        These are symbols defined as SYMBOL EQU EXTERNAL+N and exported.
        After all externals are resolved, we can compute the actual addresses
        for these aliased symbols and add them to the global table.
        """
        for mod_idx, module in enumerate(self.modules):
            for new_name, (base_ext, offset) in module.aliased_publics.items():
                # Look up the base external symbol
                if base_ext not in self.globals:
                    self.error(f"Aliased symbol '{new_name}' references undefined external '{base_ext}'")
                    continue

                base_mod_idx, base_value, base_seg_type, is_defined = self.globals[base_ext]
                if not is_defined:
                    self.error(f"Aliased symbol '{new_name}' references undefined external '{base_ext}'")
                    continue

                # The new symbol's value is base_value + offset, same segment type
                new_value = base_value + offset
                # Register the aliased symbol in globals
                if new_name in self.globals and self.globals[new_name][3]:
                    self.warning(f"Multiple definition of '{new_name}'")
                else:
                    # Use the same module index as the base external
                    self.globals[new_name] = (base_mod_idx, new_value, base_seg_type, True)

    def calculate_addresses(self):
        """Calculate base addresses for all modules."""
        # Calculate total code size
        total_code = 0
        for module in self.modules:
            module.code_base = self.code_base + total_code
            total_code += module.code_size if module.code_size else len(module.code)

        # Data follows code
        if self.data_base is None:
            self.data_base = self.code_base + total_code

        total_data = 0
        for module in self.modules:
            module.data_base = self.data_base + total_data
            total_data += module.data_size

        # Common follows data
        if self.common_base is None:
            self.common_base = self.data_base + total_data

        # Calculate total common size
        total_common = sum(self.commons.values())

        # Add __END__ symbol pointing to first free byte after all segments
        # This is an absolute address, not module-relative
        end_addr = self.common_base + total_common
        self.globals['__END__'] = (0, end_addr, ADDR_ABSOLUTE, True)

        # BSS region = COMMON area (uninitialized data, zeroed by crt0)
        self.globals['__BSS_START'] = (0, self.common_base, ADDR_ABSOLUTE, True)
        self.globals['__BSS_END'] = (0, end_addr, ADDR_ABSOLUTE, True)

    def relocate_value(self, module, value, seg_type):
        """Relocate a value based on its segment type."""
        if seg_type == ADDR_ABSOLUTE:
            return value
        elif seg_type == ADDR_PROGRAM_REL:
            return value + module.code_base
        elif seg_type == ADDR_DATA_REL:
            return value + module.data_base
        elif seg_type == ADDR_COMMON_REL:
            return value + self.common_base
        return value

    def link(self):
        """Link all loaded modules."""
        # Resolve aliased public symbols first (EQU external+offset made PUBLIC)
        # These need to be in globals before resolve_externals() checks references
        self.resolve_aliased_publics()

        if not self.resolve_externals():
            return False

        self.calculate_addresses()

        # Build output starting at origin (code_base of first module)
        # Determine output base address (lowest module address)
        self.output_base = min(m.code_base for m in self.modules)

        # Calculate total output size - must cover both CSEG and DSEG regions
        # CSEG and DSEG have separate base addresses and must be placed correctly
        total_size = 0
        for module in self.modules:
            # CSEG end address
            cseg_bytes = module.code_size if module.code_size else len(module.code)
            cseg_end = module.code_base + cseg_bytes - self.output_base
            if cseg_end > total_size:
                total_size = cseg_end

            # DSEG end address
            if module.data_size > 0:
                if self.emit_ds_zeros:
                    # Include full declared DSEG (DS directives filled with zeros)
                    dseg_end = module.data_base + module.data_size - self.output_base
                    if dseg_end > total_size:
                        total_size = dseg_end
                else:
                    # Only include actually-initialized DSEG bytes (DB/DW, not DS)
                    buffer_len = len(module.code) - module.code_start
                    initialized_dseg = buffer_len - (module.code_size if module.code_size else buffer_len)
                    if initialized_dseg > 0:
                        dseg_end = module.data_base + initialized_dseg - self.output_base
                        if dseg_end > total_size:
                            total_size = dseg_end

        self.output = bytearray(total_size)

        # Copy and relocate each module
        for module in self.modules:
            src_start = module.code_start

            # Determine CSEG size (declared or inferred from buffer)
            cseg_size = module.code_size if module.code_size else len(module.code) - src_start

            # Copy CSEG bytes to code_base
            cseg_dest = module.code_base - self.output_base
            for i in range(cseg_size):
                src_idx = src_start + i
                if src_idx < len(module.code) and cseg_dest + i < len(self.output):
                    self.output[cseg_dest + i] = module.code[src_idx]

            # Copy initialized DSEG bytes to data_base (separate from CSEG)
            if module.data_size > 0:
                buffer_len = len(module.code) - src_start
                initialized_dseg = buffer_len - cseg_size
                if initialized_dseg > 0:
                    dseg_src = src_start + cseg_size
                    dseg_dest = module.data_base - self.output_base
                    for i in range(initialized_dseg):
                        src_idx = dseg_src + i
                        if src_idx < len(module.code) and dseg_dest + i < len(self.output):
                            self.output[dseg_dest + i] = module.code[src_idx]

        # Fix up external references
        # Track which (module_index, buf_offset) pairs are resolved externally
        # so Phase 2 relocation doesn't double-apply segment bases
        resolved_external_locs = set()

        for mod_idx, module in enumerate(self.modules):
            dest_offset = module.code_base - self.output_base
            src_start = module.code_start  # For absolute ORG adjustment

            for name, refs in module.externals.items():
                # Parse "SYMBOL+N" format for expression offsets
                expr_offset = 0
                base_name = name
                if '+' in name:
                    parts = name.rsplit('+', 1)
                    base_name = parts[0]
                    try:
                        expr_offset = int(parts[1])
                    except ValueError:
                        pass  # Not a valid offset, use full name

                if base_name not in self.globals:
                    continue

                target_mod_idx, target_value, target_seg_type, _ = self.globals[base_name]
                target_module = self.modules[target_mod_idx]
                target_addr = self.relocate_value(target_module, target_value, target_seg_type)
                target_addr += expr_offset  # Add expression offset (e.g., +1 for SYMBOL+1)

                for head, ref_seg_type in refs:
                    # Mark this location as externally resolved so Phase 2 skips it
                    resolved_external_locs.add((mod_idx, head))

                    # Follow the chain and fix up each reference
                    # head is now a buffer offset, convert to output offset
                    # Use appropriate base for the reference's segment type
                    if ref_seg_type == ADDR_DATA_REL:
                        ref_dest_offset = module.data_base - self.output_base
                        seg_offset = head - module.seg_buf_start.get(ADDR_DATA_REL, 0)
                    else:
                        ref_dest_offset = dest_offset
                        seg_offset = head - src_start
                    seg_base = module.seg_buf_start.get(ref_seg_type, 0)
                    offset = seg_offset
                    visited = set()  # Prevent infinite loops
                    while offset not in visited and offset >= 0:
                        visited.add(offset)
                        abs_offset = ref_dest_offset + offset
                        if abs_offset + 1 < len(self.output) and abs_offset >= 0:
                            # Get value at this location (this is the next chain link)
                            # For ADDR_ABSOLUTE, values are absolute addresses
                            # For relative segments, values are segment-relative offsets
                            value = self.output[abs_offset] | (self.output[abs_offset + 1] << 8)

                            # Chain format: each link points to previous reference
                            # End of chain is marked by value 0
                            # Resolve this location with target address
                            self.output[abs_offset] = target_addr & 0xFF
                            self.output[abs_offset + 1] = (target_addr >> 8) & 0xFF

                            # Track for PRL relocation if target is relocatable
                            if target_seg_type in (ADDR_PROGRAM_REL, ADDR_DATA_REL, ADDR_COMMON_REL):
                                self.external_relocations.append(abs_offset)

                            if value == 0:
                                # End of chain
                                break
                            else:
                                # Follow chain to previous reference
                                # Value is segment-relative, convert to segment offset
                                offset = value
                        else:
                            break

        # Apply relocations for program-relative, data-relative, and common-relative addresses
        for mod_idx, module in enumerate(self.modules):
            src_start = module.code_start  # For absolute ORG adjustment
            for reloc_entry in module.relocations:
                # Handle both old 2-tuple and new 3-tuple format
                if len(reloc_entry) == 3:
                    buf_offset, seg_type, ref_seg_type = reloc_entry
                else:
                    buf_offset, seg_type = reloc_entry
                    ref_seg_type = ADDR_PROGRAM_REL  # Default to CSEG for compatibility

                # Skip locations already resolved by external reference fixup
                if (mod_idx, buf_offset) in resolved_external_locs:
                    continue

                # Calculate output offset based on which segment the reference is in
                if ref_seg_type == ADDR_DATA_REL:
                    # Reference is in DSEG
                    dseg_start = module.seg_buf_start.get(ADDR_DATA_REL, 0)
                    seg_offset = buf_offset - dseg_start
                    abs_offset = (module.data_base - self.output_base) + seg_offset
                else:
                    # Reference is in CSEG (or other)
                    abs_offset = (module.code_base - self.output_base) + (buf_offset - src_start)

                if abs_offset >= 0 and abs_offset + 1 < len(self.output):
                    # Read current value
                    value = self.output[abs_offset] | (self.output[abs_offset + 1] << 8)
                    # Apply relocation based on what the value points to
                    if seg_type == ADDR_PROGRAM_REL:
                        value += module.code_base
                    elif seg_type == ADDR_DATA_REL:
                        value += module.data_base
                    elif seg_type == ADDR_COMMON_REL:
                        value += self.common_base
                    # Write relocated value
                    self.output[abs_offset] = value & 0xFF
                    self.output[abs_offset + 1] = (value >> 8) & 0xFF

        return True

    def save_com(self, filename):
        """Save as CP/M .COM file."""
        # For .COM file, code loads and executes at 0x100
        # Only prepend JMP if entry point is not at 0x100
        with open(filename, 'wb') as f:
            f.write(bytes(self.output))
            # Pad to CP/M record boundary (128 bytes)
            remainder = len(self.output) % 128
            if remainder:
                f.write(bytes(128 - remainder))

    def save_hex(self, filename):
        """Save as Intel HEX format."""
        with open(filename, 'w') as f:
            addr = self.output_base
            data = bytes(self.output)

            idx = 0
            while idx < len(data):
                # Write 16 bytes per line
                line_len = min(16, len(data) - idx)
                line_data = data[idx:idx + line_len]

                # Calculate checksum
                checksum = line_len + (addr >> 8) + (addr & 0xFF) + 0  # record type 0
                checksum += sum(line_data)
                checksum = (~checksum + 1) & 0xFF

                # Write line
                hex_data = ''.join(f'{b:02X}' for b in line_data)
                f.write(f':{line_len:02X}{addr:04X}00{hex_data}{checksum:02X}\n')

                addr += line_len
                idx += line_len

            # Write EOF record
            f.write(':00000001FF\n')

    def save_prl(self, filename):
        """Save as MP/M .PRL (Page Relocatable) format.

        PRL files can be loaded at any page boundary. The relocation bitmap
        marks high bytes of 16-bit addresses that need adjustment when loaded
        at a different page than 0x100.

        Format:
        - 256-byte header (code length at offset 1-2, BSS at 4-5)
        - Code/data (code_length bytes)
        - Relocation bitmap ((code_length + 7) / 8 bytes)
        """
        code_length = len(self.output)

        # Calculate total BSS (uninitialized data) size from all modules
        # BSS = declared DSEG size - initialized DSEG bytes actually emitted
        # Initialized DSEG bytes = total buffer - code_start - CSEG size
        bss_size = 0
        for module in self.modules:
            actual_bytes = len(module.code) - module.code_start
            cseg_size = module.code_size if module.code_size else 0
            initialized_dseg = actual_bytes - cseg_size
            uninitialized_dseg = module.data_size - initialized_dseg
            if uninitialized_dseg > 0:
                bss_size += uninitialized_dseg

        # Build relocation bitmap - one bit per byte of code
        # Bit is set if corresponding byte is a HIGH byte of relocatable address
        bitmap_size = (code_length + 7) // 8
        bitmap = bytearray(bitmap_size)

        # Collect all relocation high-byte offsets from all modules
        for module in self.modules:
            dest_offset = module.code_base - self.output_base
            src_start = module.code_start

            for reloc_entry in module.relocations:
                # Handle both old 2-tuple and new 3-tuple format
                if len(reloc_entry) == 3:
                    buf_offset, seg_type, ref_seg_type = reloc_entry
                else:
                    buf_offset, seg_type = reloc_entry
                    ref_seg_type = ADDR_PROGRAM_REL

                # Calculate output offset based on which segment the reference is in
                if ref_seg_type == ADDR_DATA_REL:
                    dseg_start = module.seg_buf_start.get(ADDR_DATA_REL, 0)
                    seg_offset = buf_offset - dseg_start
                    abs_offset = (module.data_base - self.output_base) + seg_offset
                else:
                    abs_offset = dest_offset + (buf_offset - src_start)

                high_byte_offset = abs_offset + 1

                if 0 <= high_byte_offset < code_length:
                    # Set bit in bitmap
                    # Bit 7 of byte 0 = code byte 0, bit 6 = code byte 1, etc.
                    byte_idx = high_byte_offset // 8
                    bit_idx = 7 - (high_byte_offset % 8)
                    bitmap[byte_idx] |= (1 << bit_idx)

        # Also include external relocations (resolved refs to CSEG symbols)
        for abs_offset in self.external_relocations:
            high_byte_offset = abs_offset + 1
            if 0 <= high_byte_offset < code_length:
                byte_idx = high_byte_offset // 8
                bit_idx = 7 - (high_byte_offset % 8)
                bitmap[byte_idx] |= (1 << bit_idx)

        # Build header (256 bytes)
        # MP/M II PRL format (verified from PRLCM.PLM):
        # Byte 0: Type/reserved (0)
        # Bytes 1-2: Code length (16-bit little-endian)
        # Byte 3: Reserved (0)
        # Bytes 4-5: BSS/extra size (optional)
        # Bytes 6-255: Reserved (0)
        header = bytearray(256)
        header[0] = 0  # Type/reserved
        header[1] = code_length & 0xFF  # Code length low byte
        header[2] = (code_length >> 8) & 0xFF  # Code length high byte
        header[3] = 0  # Reserved
        header[4] = bss_size & 0xFF  # BSS size low byte
        header[5] = (bss_size >> 8) & 0xFF  # BSS size high byte
        header[6] = 0  # Always 0
        header[7] = 0  # Load address low (0 for PRL)
        header[8] = 0  # Load address high (0 for PRL)
        # Bytes 9-255 remain 0

        with open(filename, 'wb') as f:
            f.write(bytes(header))
            f.write(bytes(self.output))
            f.write(bytes(bitmap))
            # Pad to CP/M record boundary (128 bytes)
            total_size = 256 + code_length + bitmap_size
            remainder = total_size % 128
            if remainder:
                f.write(bytes(128 - remainder))

    def save_sym(self, filename):
        """Save symbol table file (.SYM) compatible with SID/ZSID debuggers.

        Format: ADDR NAME (one per line, LF endings)
        Sorted alphabetically by symbol name.
        """
        # Build list of (name, address) for all defined globals
        symbols = []
        for name, (mod_idx, value, seg_type, is_defined) in self.globals.items():
            if is_defined:
                module = self.modules[mod_idx]
                addr = self.relocate_value(module, value, seg_type)
                symbols.append((name, addr))

        # Sort alphabetically by symbol name (DRI convention)
        symbols.sort(key=lambda x: x[0])

        with open(filename, 'w') as f:
            # SID .SYM format: ADDR NAME (one per line)
            # Use LF line endings - cpmemu converts to CR-LF in text mode
            for name, addr in symbols:
                f.write(f"{addr:04X} {name}\n")


def main():
    parser = argparse.ArgumentParser(description='ul80 - LINK-80 compatible linker')
    parser.add_argument('-v', '--version', action='version', version=f'%(prog)s {__version__}')
    parser.add_argument('inputs', nargs='+', help='Input .REL and .LIB files')
    parser.add_argument('-o', '--output', help='Output file (default: first input with .com)')
    parser.add_argument('-x', '--hex', action='store_true', help='Output Intel HEX format')
    parser.add_argument('--prl', action='store_true', help='Output MP/M .PRL (Page Relocatable) format')
    parser.add_argument('--no-ds-zeros', action='store_true',
                       help='Do not emit zeros for DS (reserve space) directives (default: emit zeros)')
    parser.add_argument('-s', '--sym', action='store_true', help='Generate .SYM symbol file')
    parser.add_argument('-S', '--sym-file', metavar='FILE', help='Generate .SYM symbol file with specified name')
    parser.add_argument('-p', '--origin', type=lambda x: int(x, 16) if not x.startswith(('0x', '0X', '0o', '0O', '0b', '0B')) else int(x, 0), default=None,
                       help='Program origin as hex (e.g., E000, 0xE000, default: 0 for PRL, 100 for COM)')

    args = parser.parse_args()

    linker = Linker()
    # Default origin: 0 for PRL/SPR (page relocatable), 0x100 for COM (CP/M TPA)
    if args.origin is None:
        args.origin = 0 if args.prl else 0x100
    linker.code_base = args.origin

    # Emit zeros for DS directives (default: True)
    if args.no_ds_zeros:
        linker.emit_ds_zeros = False

    # Separate .rel and .lib files
    rel_files = []
    lib_files = []
    for filename in args.inputs:
        if not os.path.exists(filename):
            print(f"Error: File not found: {filename}", file=sys.stderr)
            sys.exit(1)
        ext = Path(filename).suffix.lower()
        if ext == '.lib':
            lib_files.append(filename)
        else:
            rel_files.append(filename)

    # Load all .rel files first
    for filename in rel_files:
        if not linker.load_rel(filename):
            print(f"Error loading {filename}", file=sys.stderr)
            sys.exit(1)

    # Load libraries
    libraries = []
    for filename in lib_files:
        try:
            lib = Library.load(filename)
            libraries.append((filename, lib))
        except LibraryError as e:
            print(f"Error loading library {filename}: {e}", file=sys.stderr)
            sys.exit(1)

    # Resolve undefined symbols from libraries
    # Keep searching until no more symbols can be resolved
    modules_loaded = set()  # Track which library modules we've already loaded
    while libraries:
        undefined = linker.get_undefined_symbols()
        if not undefined:
            break

        resolved_any = False
        for symbol in list(undefined):
            # Search libraries for this symbol
            for lib_filename, lib in libraries:
                module_name = lib.find_module_for_symbol(symbol)
                if module_name:
                    # Check if we already loaded this module
                    lib_mod_key = (lib_filename, module_name)
                    if lib_mod_key in modules_loaded:
                        continue

                    # Get the module and load its REL data
                    lib_module = lib.get_module(module_name)
                    if lib_module:
                        linker.load_rel_data(module_name, lib_module.data)
                        modules_loaded.add(lib_mod_key)
                        resolved_any = True
                        break
            if resolved_any:
                break  # Restart the search with updated undefined symbols

        if not resolved_any:
            # No more symbols can be resolved from libraries
            break

    # Link
    if not linker.link():
        for err in linker.errors:
            print(err, file=sys.stderr)
        sys.exit(1)

    # Determine output filename
    if args.output:
        output_path = args.output
    else:
        if args.hex:
            ext = '.hex'
        elif args.prl:
            ext = '.prl'
        else:
            ext = '.com'
        output_path = Path(args.inputs[0]).with_suffix(ext)

    # Save output
    if args.hex:
        linker.save_hex(str(output_path))
    elif args.prl:
        linker.save_prl(str(output_path))
    else:
        linker.save_com(str(output_path))

    # Save symbol file if requested
    if args.sym or args.sym_file:
        if args.sym_file:
            sym_path = Path(args.sym_file)
        else:
            sym_path = Path(output_path).with_suffix('.sym')
        linker.save_sym(str(sym_path))
        print(f"Symbol file -> {sym_path}")

    # Report warnings
    for warn in linker.warnings:
        print(warn, file=sys.stderr)

    print(f"Linked -> {output_path}")
    print(f"  Modules: {len(linker.modules)}")
    print(f"  Global symbols: {len(linker.globals)}")

    sys.exit(0)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
"""
ucref80 - Cross-reference utility for MACRO-80 assembly files.

Generates a cross-reference listing showing where symbols are defined
and referenced in assembly source files.

Usage: ucref80 [-o output.txt] file1.mac [file2.mac ...]
"""

import sys
import os
import argparse
import re
from collections import defaultdict

from um80 import __version__


class Symbol:
    """Symbol cross-reference entry."""
    def __init__(self, name):
        self.name = name
        self.definitions = []  # List of (filename, line_num)
        self.references = []   # List of (filename, line_num)
        self.is_external = False
        self.is_public = False


class CrossReference:
    """Cross-reference generator."""

    def __init__(self):
        self.symbols = {}  # name -> Symbol
        self.current_file = ""
        self.files_processed = []

        # Patterns for parsing
        self.label_pattern = re.compile(r'^([A-Za-z_$@?][A-Za-z0-9_$@?]*)\s*:', re.IGNORECASE)
        self.equ_pattern = re.compile(r'^([A-Za-z_$@?][A-Za-z0-9_$@?]*)\s+(equ|set|defl)\s+', re.IGNORECASE)
        self.public_pattern = re.compile(r'^\s*public\s+(.+)', re.IGNORECASE)
        self.extrn_pattern = re.compile(r'^\s*extrn\s+(.+)', re.IGNORECASE)
        self.macro_pattern = re.compile(r'^([A-Za-z_$@?][A-Za-z0-9_$@?]*)\s+macro\b', re.IGNORECASE)

        # Reserved words that shouldn't be treated as symbols
        self.reserved = {
            # Directives
            'ORG', 'EQU', 'SET', 'DEFL', 'DB', 'DW', 'DS', 'DC', 'DEFB', 'DEFW', 'DEFS',
            'IF', 'ELSE', 'ENDIF', 'IFT', 'IFF', 'COND', 'ENDC',
            'IFDEF', 'IFNDEF', 'IFB', 'IFNB', 'IFIDN', 'IFDIF',
            'MACRO', 'ENDM', 'LOCAL', 'REPT', 'IRP', 'IRPC', 'EXITM',
            'PUBLIC', 'EXTRN', 'EXTERN', 'GLOBAL', 'ENTRY', 'EXT', 'NAME',
            'CSEG', 'DSEG', 'ASEG', 'COMMON',
            'END', 'TITLE', 'SUBTTL', 'PAGE', 'EJECT',
            '.Z80', '.8080', '.8085', '.PHASE', '.DEPHASE',
            '.LIST', '.XLIST', '.SALL', '.LALL', '.XALL', '.SFCOND', '.LFCOND',
            '.PRINTX', '.RADIX', '.REQUEST', '.COMMENT',
            'INCLUDE', 'INCBIN',
            # 8080/Z80 Instructions
            'MOV', 'MVI', 'LXI', 'LDA', 'STA', 'LHLD', 'SHLD', 'LDAX', 'STAX',
            'XCHG', 'PUSH', 'POP', 'XTHL', 'SPHL', 'PCHL',
            'ADD', 'ADC', 'SUB', 'SBB', 'ANA', 'XRA', 'ORA', 'CMP',
            'ADI', 'ACI', 'SUI', 'SBI', 'ANI', 'XRI', 'ORI', 'CPI',
            'INR', 'DCR', 'INX', 'DCX', 'DAD', 'DAA',
            'RLC', 'RRC', 'RAL', 'RAR', 'CMA', 'CMC', 'STC',
            'JMP', 'JC', 'JNC', 'JZ', 'JNZ', 'JP', 'JM', 'JPE', 'JPO',
            'CALL', 'CC', 'CNC', 'CZ', 'CNZ', 'CP', 'CM', 'CPE', 'CPO',
            'RET', 'RC', 'RNC', 'RZ', 'RNZ', 'RP', 'RM', 'RPE', 'RPO',
            'RST', 'EI', 'DI', 'NOP', 'HLT', 'RIM', 'SIM',
            'IN', 'OUT',
            # Z80 extensions
            'LD', 'EX', 'EXX', 'DJNZ', 'JR', 'JP', 'LDIR', 'LDDR', 'CPIR', 'CPDR',
            'NEG', 'CPL', 'SCF', 'CCF', 'HALT', 'RETI', 'RETN',
            'BIT', 'SET', 'RES', 'RL', 'RR', 'SLA', 'SRA', 'SRL', 'RLD', 'RRD',
            'IM', 'OTIR', 'OTDR', 'INIR', 'INDR', 'OUTI', 'OUTD', 'INI', 'IND',
            # Registers
            'A', 'B', 'C', 'D', 'E', 'H', 'L', 'M', 'SP', 'PSW',
            'AF', 'BC', 'DE', 'HL', 'IX', 'IY', 'I', 'R',
            # Conditions
            'NZ', 'NC', 'PO', 'PE', 'P',
        }

    def get_symbol(self, name):
        """Get or create a symbol entry."""
        name = name.upper()
        if name not in self.symbols:
            self.symbols[name] = Symbol(name)
        return self.symbols[name]

    def add_definition(self, name, filename, line_num):
        """Record a symbol definition."""
        sym = self.get_symbol(name)
        sym.definitions.append((filename, line_num))

    def add_reference(self, name, filename, line_num):
        """Record a symbol reference."""
        sym = self.get_symbol(name)
        sym.references.append((filename, line_num))

    def is_reserved(self, name):
        """Check if a name is a reserved word."""
        return name.upper() in self.reserved

    def extract_symbols(self, text):
        """Extract potential symbol names from text."""
        # Remove strings
        text = re.sub(r"'[^']*'", "", text)
        text = re.sub(r'"[^"]*"', "", text)

        # Find all identifier-like tokens
        symbols = re.findall(r'[A-Za-z_$@?][A-Za-z0-9_$@?]*', text)
        return [s for s in symbols if not self.is_reserved(s)]

    def process_line(self, line, filename, line_num):
        """Process a single line of assembly source."""
        # Strip CR/LF and get original line
        line = line.rstrip('\r\n')

        # Remove comments
        if ';' in line:
            code_part = line.split(';')[0]
        else:
            code_part = line

        if not code_part.strip():
            return

        # Check for label definition (label at start of line with colon)
        match = self.label_pattern.match(code_part)
        if match:
            label = match.group(1)
            if not self.is_reserved(label):
                self.add_definition(label, filename, line_num)
            # Continue processing rest of line for references
            code_part = code_part[match.end():]

        # Check for EQU/SET/DEFL definition (symbol before EQU)
        match = self.equ_pattern.match(code_part)
        if match:
            symbol = match.group(1)
            if not self.is_reserved(symbol):
                self.add_definition(symbol, filename, line_num)
            # The expression after EQU may contain references
            code_part = code_part[match.end():]
            for sym in self.extract_symbols(code_part):
                self.add_reference(sym, filename, line_num)
            return

        # Check for MACRO definition
        match = self.macro_pattern.match(code_part)
        if match:
            macro_name = match.group(1)
            if not self.is_reserved(macro_name):
                self.add_definition(macro_name, filename, line_num)
            return

        # Check for PUBLIC declaration
        match = self.public_pattern.match(code_part)
        if match:
            symbols_text = match.group(1)
            for sym in self.extract_symbols(symbols_text):
                self.get_symbol(sym).is_public = True
            return

        # Check for EXTRN declaration
        match = self.extrn_pattern.match(code_part)
        if match:
            symbols_text = match.group(1)
            for sym in self.extract_symbols(symbols_text):
                s = self.get_symbol(sym)
                s.is_external = True
            return

        # Skip directives that contain text, not code
        upper_code = code_part.upper().strip()
        skip_directives = ['TITLE', 'SUBTTL', '.PRINTX', 'NAME', 'PAGE', 'EJECT',
                          '.COMMENT', '.SALL', '.LALL', '.XALL', '.LIST', '.XLIST',
                          '.SFCOND', '.LFCOND', '.RADIX', '.REQUEST']
        for directive in skip_directives:
            if upper_code.startswith(directive):
                return

        # For all other lines, extract symbol references
        # Skip the opcode/directive at the start
        parts = code_part.split(None, 1)
        if len(parts) > 1:
            operands = parts[1]
            for sym in self.extract_symbols(operands):
                self.add_reference(sym, filename, line_num)
        elif len(parts) == 1:
            # Could be a macro invocation
            potential_macro = parts[0]
            if not self.is_reserved(potential_macro):
                # Check if it's defined as a macro
                sym = self.symbols.get(potential_macro.upper())
                if sym and sym.definitions:
                    self.add_reference(potential_macro, filename, line_num)

    def process_file(self, filename):
        """Process an assembly source file."""
        self.current_file = os.path.basename(filename)
        self.files_processed.append(self.current_file)

        try:
            with open(filename, 'rb') as f:
                data = f.read()
        except IOError as e:
            print(f"Error reading {filename}: {e}", file=sys.stderr)
            return False

        # Handle CP/M format (strip ^Z and everything after)
        if b'\x1a' in data:
            data = data[:data.index(b'\x1a')]

        # Decode and split into lines
        text = data.decode('utf-8', errors='replace')

        lines = text.replace('\r\n', '\n').replace('\r', '\n').split('\n')

        in_macro = False
        macro_depth = 0

        for line_num, line in enumerate(lines, 1):
            # Track macro depth (don't process macro bodies for references)
            upper_line = line.upper().strip()
            if re.match(r'^\w+\s+MACRO\b', upper_line) or upper_line.startswith('MACRO '):
                in_macro = True
                macro_depth += 1
            elif upper_line.strip() == 'ENDM':
                macro_depth -= 1
                if macro_depth <= 0:
                    in_macro = False
                    macro_depth = 0
                continue

            # Still process definitions inside macros, but skip reference tracking
            if in_macro:
                # Only look for definitions
                match = self.label_pattern.match(line)
                if match:
                    label = match.group(1)
                    if not self.is_reserved(label):
                        self.add_definition(label, self.current_file, line_num)
                continue

            self.process_line(line, self.current_file, line_num)

        return True

    def generate_report(self, output_file=None):
        """Generate the cross-reference report."""
        lines = []

        # Header
        lines.append("=" * 78)
        lines.append("CROSS REFERENCE LISTING")
        lines.append("=" * 78)
        lines.append("")
        lines.append(f"Files: {', '.join(self.files_processed)}")
        lines.append(f"Symbols: {len(self.symbols)}")
        lines.append("")
        lines.append("-" * 78)
        lines.append("")

        # Sort symbols alphabetically
        sorted_symbols = sorted(self.symbols.values(), key=lambda s: s.name)

        for sym in sorted_symbols:
            # Skip symbols with no definitions and no references
            if not sym.definitions and not sym.references:
                continue

            # Symbol name with flags
            flags = []
            if sym.is_public:
                flags.append("PUBLIC")
            if sym.is_external:
                flags.append("EXTRN")

            flag_str = f" ({', '.join(flags)})" if flags else ""
            lines.append(f"{sym.name}{flag_str}")

            # Definitions
            if sym.definitions:
                def_strs = []
                for filename, line_num in sym.definitions:
                    def_strs.append(f"{filename}:{line_num}")
                lines.append(f"  Defined:    {', '.join(def_strs)}")
            elif sym.is_external:
                lines.append(f"  Defined:    (external)")
            else:
                lines.append(f"  Defined:    ** UNDEFINED **")

            # References
            if sym.references:
                # Group references by file
                refs_by_file = defaultdict(list)
                for filename, line_num in sym.references:
                    refs_by_file[filename].append(line_num)

                ref_parts = []
                for filename in sorted(refs_by_file.keys()):
                    line_nums = sorted(set(refs_by_file[filename]))
                    # Format line numbers, wrapping if needed
                    nums_str = ', '.join(str(n) for n in line_nums)
                    ref_parts.append(f"{filename}: {nums_str}")

                # Wrap long reference lists
                ref_str = '; '.join(ref_parts)
                if len(ref_str) > 60:
                    lines.append(f"  References:")
                    for part in ref_parts:
                        lines.append(f"    {part}")
                else:
                    lines.append(f"  References: {ref_str}")
            else:
                lines.append(f"  References: (none)")

            lines.append("")

        # Summary
        lines.append("-" * 78)
        lines.append("")

        # Count statistics
        defined = sum(1 for s in sorted_symbols if s.definitions)
        external = sum(1 for s in sorted_symbols if s.is_external)
        public = sum(1 for s in sorted_symbols if s.is_public)
        undefined = sum(1 for s in sorted_symbols
                       if not s.definitions and not s.is_external and s.references)

        lines.append("SUMMARY")
        lines.append(f"  Total symbols:     {len(sorted_symbols)}")
        lines.append(f"  Defined:           {defined}")
        lines.append(f"  External:          {external}")
        lines.append(f"  Public:            {public}")
        if undefined:
            lines.append(f"  Undefined:         {undefined}")
        lines.append("")

        # Undefined symbols listing
        undefined_syms = [s for s in sorted_symbols
                        if not s.definitions and not s.is_external and s.references]
        if undefined_syms:
            lines.append("UNDEFINED SYMBOLS:")
            for sym in undefined_syms:
                lines.append(f"  {sym.name}")
            lines.append("")

        report = '\n'.join(lines)

        if output_file:
            with open(output_file, 'w') as f:
                f.write(report)
            print(f"Cross-reference written to {output_file}")
        else:
            print(report)

        return report


def main():
    parser = argparse.ArgumentParser(
        description='ucref80 - Generate cross-reference listing for MACRO-80 assembly files')
    parser.add_argument('-v', '--version', action='version', version=f'%(prog)s {__version__}')
    parser.add_argument('inputs', nargs='+', help='Input .MAC files')
    parser.add_argument('-o', '--output', help='Output file (default: stdout)')

    args = parser.parse_args()

    # Check input files exist
    for filename in args.inputs:
        if not os.path.exists(filename):
            print(f"Error: File not found: {filename}", file=sys.stderr)
            sys.exit(1)

    # Process files
    cref = CrossReference()

    for filename in args.inputs:
        print(f"Processing {filename}...", file=sys.stderr)
        if not cref.process_file(filename):
            sys.exit(1)

    # Generate report
    cref.generate_report(args.output)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
"""
ud80 - 8080/Z80 Disassembler for CP/M .COM files.

Produces .MAC output compatible with um80 assembler.

Usage: ud80 [-z] [-o output.mac] [-e entry] [-l label] [-d range] [-t range] input.com

Options:
    -z, --z80       Z80 mode: use Zilog mnemonics and decode Z80-only opcodes
                    (CB, DD, ED, FD prefixes including undocumented instructions)
    -o, --output    Output file (default: input.mac)
    -e, --entry     Entry point to trace as code (e.g., -e 1234 or -e 1234,myproc)
                    Can specify optional label name after comma
    -l, --label     Add label without tracing as code (e.g., -l 1A00 or -l 1A00,buffer)
                    Useful for data labels or unreached code
    -d, --data      Force address range to be data, not code (e.g., -d 1A00-1A7F)
                    Can be repeated for multiple ranges
    -t, --table     Address table range, output as DW with labels (e.g., -t 0103-0120)
                    Can be repeated for multiple ranges
    --org           Origin address (default: 0100 for CP/M)
"""

import sys
import os
import argparse
from collections import defaultdict

from um80 import __version__

# 8080 instruction table: opcode -> (mnemonic, size, operand_type)
# Operand types:
#   None - no operands
#   'r' - single register in bits 3-5
#   's' - single register in bits 0-2
#   'rp' - register pair in bits 4-5
#   'rp2' - register pair for PUSH/POP (includes PSW)
#   'd8' - 8-bit immediate
#   'd16' - 16-bit immediate/address
#   'a16' - 16-bit address (for jumps/calls)
#   'rst' - RST vector in bits 3-5

REGS = ['B', 'C', 'D', 'E', 'H', 'L', 'M', 'A']
REGPAIRS = ['BC', 'DE', 'HL', 'SP']
REGPAIRS_PUSH = ['BC', 'DE', 'HL', 'PSW']
CONDITIONS = ['NZ', 'Z', 'NC', 'C', 'PO', 'PE', 'P', 'M']

# Build instruction table
OPCODES = {}

# No operand instructions
NO_OP = {
    0x00: 'NOP', 0x07: 'RLC', 0x0F: 'RRC', 0x17: 'RAL', 0x1F: 'RAR',
    0x27: 'DAA', 0x2F: 'CMA', 0x37: 'STC', 0x3F: 'CMC', 0x76: 'HLT',
    0xC9: 'RET', 0xE9: 'PCHL', 0xF9: 'SPHL', 0xEB: 'XCHG', 0xE3: 'XTHL',
    0xF3: 'DI', 0xFB: 'EI',
}
for op, mn in NO_OP.items():
    OPCODES[op] = (mn, 1, None)

# MOV r,r' (01 ddd sss) - but 76 is HLT
for d in range(8):
    for s in range(8):
        op = 0x40 | (d << 3) | s
        if op != 0x76:  # HLT
            OPCODES[op] = ('MOV', 1, ('r', 's'))

# MVI r,d8 (00 rrr 110)
for r in range(8):
    op = 0x06 | (r << 3)
    OPCODES[op] = ('MVI', 2, ('r', 'd8'))

# LXI rp,d16 (00 rp0 001)
for rp in range(4):
    op = 0x01 | (rp << 4)
    OPCODES[op] = ('LXI', 3, ('rp', 'd16'))

# INR r (00 rrr 100)
for r in range(8):
    op = 0x04 | (r << 3)
    OPCODES[op] = ('INR', 1, ('r',))

# DCR r (00 rrr 101)
for r in range(8):
    op = 0x05 | (r << 3)
    OPCODES[op] = ('DCR', 1, ('r',))

# INX rp (00 rp0 011)
for rp in range(4):
    op = 0x03 | (rp << 4)
    OPCODES[op] = ('INX', 1, ('rp',))

# DCX rp (00 rp1 011)
for rp in range(4):
    op = 0x0B | (rp << 4)
    OPCODES[op] = ('DCX', 1, ('rp',))

# DAD rp (00 rp1 001)
for rp in range(4):
    op = 0x09 | (rp << 4)
    OPCODES[op] = ('DAD', 1, ('rp',))

# LDAX rp (00 rp1 010) - only BC,DE
for rp in range(2):
    op = 0x0A | (rp << 4)
    OPCODES[op] = ('LDAX', 1, ('rp',))

# STAX rp (00 rp0 010) - only BC,DE
for rp in range(2):
    op = 0x02 | (rp << 4)
    OPCODES[op] = ('STAX', 1, ('rp',))

# PUSH rp (11 rp0 101)
for rp in range(4):
    op = 0xC5 | (rp << 4)
    OPCODES[op] = ('PUSH', 1, ('rp2',))

# POP rp (11 rp0 001)
for rp in range(4):
    op = 0xC1 | (rp << 4)
    OPCODES[op] = ('POP', 1, ('rp2',))

# ALU r (10 ooo rrr)
ALU_OPS = ['ADD', 'ADC', 'SUB', 'SBB', 'ANA', 'XRA', 'ORA', 'CMP']
for alu in range(8):
    for r in range(8):
        op = 0x80 | (alu << 3) | r
        OPCODES[op] = (ALU_OPS[alu], 1, ('s',))

# ALU immediate (11 ooo 110)
ALU_IMM = ['ADI', 'ACI', 'SUI', 'SBI', 'ANI', 'XRI', 'ORI', 'CPI']
for alu in range(8):
    op = 0xC6 | (alu << 3)
    OPCODES[op] = (ALU_IMM[alu], 2, ('d8',))

# Conditional returns (11 ccc 000)
for c in range(8):
    op = 0xC0 | (c << 3)
    mn = 'R' + CONDITIONS[c]
    OPCODES[op] = (mn, 1, None)

# Conditional jumps (11 ccc 010)
for c in range(8):
    op = 0xC2 | (c << 3)
    mn = 'J' + CONDITIONS[c]
    OPCODES[op] = (mn, 3, ('a16',))

# Conditional calls (11 ccc 100)
for c in range(8):
    op = 0xC4 | (c << 3)
    mn = 'C' + CONDITIONS[c]
    OPCODES[op] = (mn, 3, ('a16',))

# JMP
OPCODES[0xC3] = ('JMP', 3, ('a16',))

# CALL
OPCODES[0xCD] = ('CALL', 3, ('a16',))

# RST n (11 nnn 111)
for n in range(8):
    op = 0xC7 | (n << 3)
    OPCODES[op] = ('RST', 1, ('rst',))

# Memory reference
OPCODES[0x3A] = ('LDA', 3, ('a16',))
OPCODES[0x32] = ('STA', 3, ('a16',))
OPCODES[0x2A] = ('LHLD', 3, ('a16',))
OPCODES[0x22] = ('SHLD', 3, ('a16',))

# I/O
OPCODES[0xDB] = ('IN', 2, ('d8',))
OPCODES[0xD3] = ('OUT', 2, ('d8',))

# Z80 extended opcodes
# Z80 register names for Zilog syntax
Z80_REGS = ['B', 'C', 'D', 'E', 'H', 'L', '(HL)', 'A']
Z80_REGPAIRS = ['BC', 'DE', 'HL', 'SP']
Z80_REGPAIRS_AF = ['BC', 'DE', 'HL', 'AF']
Z80_CONDITIONS = ['NZ', 'Z', 'NC', 'C', 'PO', 'PE', 'P', 'M']

# CB-prefix opcodes: bit operations
CB_OPCODES = {}
CB_OPS = ['RLC', 'RRC', 'RL', 'RR', 'SLA', 'SRA', 'SLL', 'SRL']
for op_idx in range(8):
    for r in range(8):
        opcode = (op_idx << 3) | r
        CB_OPCODES[opcode] = (CB_OPS[op_idx], 2, ('cb_r',))

# BIT n,r (01 nnn rrr)
for n in range(8):
    for r in range(8):
        opcode = 0x40 | (n << 3) | r
        CB_OPCODES[opcode] = ('BIT', 2, ('bit_n', 'cb_r'))

# RES n,r (10 nnn rrr)
for n in range(8):
    for r in range(8):
        opcode = 0x80 | (n << 3) | r
        CB_OPCODES[opcode] = ('RES', 2, ('bit_n', 'cb_r'))

# SET n,r (11 nnn rrr)
for n in range(8):
    for r in range(8):
        opcode = 0xC0 | (n << 3) | r
        CB_OPCODES[opcode] = ('SET', 2, ('bit_n', 'cb_r'))

# ED-prefix opcodes
ED_OPCODES = {}

# IN r,(C) / OUT (C),r
for r in range(8):
    if r != 6:  # 6 is special
        ED_OPCODES[0x40 | (r << 3)] = ('IN', 2, ('ed_r', '(C)'))
        ED_OPCODES[0x41 | (r << 3)] = ('OUT', 2, ('(C)', 'ed_r'))
ED_OPCODES[0x70] = ('IN', 2, ('(C)',))  # IN (C) - result discarded

# SBC HL,rp / ADC HL,rp
for rp in range(4):
    ED_OPCODES[0x42 | (rp << 4)] = ('SBC', 2, ('HL', 'ed_rp'))
    ED_OPCODES[0x4A | (rp << 4)] = ('ADC', 2, ('HL', 'ed_rp'))

# LD (nn),rp / LD rp,(nn)
for rp in range(4):
    ED_OPCODES[0x43 | (rp << 4)] = ('LD', 4, ('(a16)', 'ed_rp'))
    ED_OPCODES[0x4B | (rp << 4)] = ('LD', 4, ('ed_rp', '(a16)'))

# NEG (all map to same instruction)
for i in range(8):
    ED_OPCODES[0x44 | (i << 3)] = ('NEG', 2, None)

# RETN / RETI
ED_OPCODES[0x45] = ('RETN', 2, None)
ED_OPCODES[0x4D] = ('RETI', 2, None)
ED_OPCODES[0x55] = ('RETN', 2, None)
ED_OPCODES[0x5D] = ('RETN', 2, None)
ED_OPCODES[0x65] = ('RETN', 2, None)
ED_OPCODES[0x6D] = ('RETN', 2, None)
ED_OPCODES[0x75] = ('RETN', 2, None)
ED_OPCODES[0x7D] = ('RETN', 2, None)

# IM 0/1/2
ED_OPCODES[0x46] = ('IM', 2, ('im0',))
ED_OPCODES[0x56] = ('IM', 2, ('im1',))
ED_OPCODES[0x5E] = ('IM', 2, ('im2',))
ED_OPCODES[0x4E] = ('IM', 2, ('im0',))
ED_OPCODES[0x66] = ('IM', 2, ('im0',))
ED_OPCODES[0x6E] = ('IM', 2, ('im0',))
ED_OPCODES[0x76] = ('IM', 2, ('im1',))
ED_OPCODES[0x7E] = ('IM', 2, ('im2',))

# Special registers
ED_OPCODES[0x47] = ('LD', 2, ('I', 'A'))
ED_OPCODES[0x4F] = ('LD', 2, ('R', 'A'))
ED_OPCODES[0x57] = ('LD', 2, ('A', 'I'))
ED_OPCODES[0x5F] = ('LD', 2, ('A', 'R'))

# Rotate BCD
ED_OPCODES[0x67] = ('RRD', 2, None)
ED_OPCODES[0x6F] = ('RLD', 2, None)

# Block operations
ED_OPCODES[0xA0] = ('LDI', 2, None)
ED_OPCODES[0xA1] = ('CPI', 2, None)
ED_OPCODES[0xA2] = ('INI', 2, None)
ED_OPCODES[0xA3] = ('OUTI', 2, None)
ED_OPCODES[0xA8] = ('LDD', 2, None)
ED_OPCODES[0xA9] = ('CPD', 2, None)
ED_OPCODES[0xAA] = ('IND', 2, None)
ED_OPCODES[0xAB] = ('OUTD', 2, None)
ED_OPCODES[0xB0] = ('LDIR', 2, None)
ED_OPCODES[0xB1] = ('CPIR', 2, None)
ED_OPCODES[0xB2] = ('INIR', 2, None)
ED_OPCODES[0xB3] = ('OTIR', 2, None)
ED_OPCODES[0xB8] = ('LDDR', 2, None)
ED_OPCODES[0xB9] = ('CPDR', 2, None)
ED_OPCODES[0xBA] = ('INDR', 2, None)
ED_OPCODES[0xBB] = ('OTDR', 2, None)

# DD/FD prefix - index register operations
# These modify HL->IX/IY, H->IXH/IYH, L->IXL/IYL in certain instructions
# We'll handle this dynamically in decode_instruction

# Z80-only base opcodes (8080 treats these as undocumented NOPs)
Z80_ONLY_OPCODES = {
    0x08: ('EX', 1, ('af_af',)),      # EX AF,AF'
    0x10: ('DJNZ', 2, ('rel8',)),     # DJNZ d
    0x18: ('JR', 2, ('rel8',)),       # JR d
    0x20: ('JR', 2, ('NZ', 'rel8')),  # JR NZ,d
    0x28: ('JR', 2, ('Z', 'rel8')),   # JR Z,d
    0x30: ('JR', 2, ('NC', 'rel8')),  # JR NC,d
    0x38: ('JR', 2, ('C', 'rel8')),   # JR C,d
}


def format_hex8(val):
    """Format 8-bit value as hex with leading 0 if needed."""
    s = f'{val:02X}H'
    if s[0].isalpha():
        s = '0' + s
    return s


def format_hex16(val):
    """Format 16-bit value as hex with leading 0 if needed."""
    s = f'{val:04X}H'
    if s[0].isalpha():
        s = '0' + s
    return s


class Disassembler:
    """8080/Z80 Disassembler."""

    def __init__(self, data, org=0x0100, z80_mode=False):
        self.data = data
        self.org = org
        self.end = org + len(data)
        self.z80_mode = z80_mode

        # Analysis results
        self.labels = {}  # addr -> label name
        self.code_addrs = set()  # Addresses that are code
        self.data_addrs = set()  # Addresses that are data
        self.table_addrs = set()  # Addresses that are address tables (DW)
        self.refs_from = defaultdict(set)  # addr -> set of addresses that reference it
        self.refs_to = defaultdict(set)  # addr -> set of addresses it references
        self.entry_points = set()  # Known entry points
        self.strings = {}  # addr -> string content
        self.post_jump_gaps = []  # List of (start, end) ranges after unconditional jumps
        self.detected_tables = set()  # Auto-detected address table addresses
        self.user_labels = {}  # addr -> user-specified label name
        self.dc_string_addrs = set()  # Addresses in high-bit terminated string regions
        self.da_string_addrs = set()  # Addresses in ASCII string regions
        self.da_string_len = {}  # addr -> fixed length (0 = null-terminated)

        # Output
        self.output_lines = []

    def byte_at(self, addr):
        """Get byte at address."""
        if addr < self.org or addr >= self.end:
            return None
        return self.data[addr - self.org]

    def word_at(self, addr):
        """Get 16-bit word at address (little endian)."""
        lo = self.byte_at(addr)
        hi = self.byte_at(addr + 1)
        if lo is None or hi is None:
            return None
        return lo | (hi << 8)

    def decode_instruction(self, addr):
        """Decode instruction at address. Returns (mnemonic, size, operands, target)."""
        opcode = self.byte_at(addr)
        if opcode is None:
            return None

        # Z80 mode: handle prefix bytes and Z80-only opcodes
        if self.z80_mode:
            if opcode == 0xCB:
                return self.decode_cb_instruction(addr)
            elif opcode == 0xED:
                return self.decode_ed_instruction(addr)
            elif opcode == 0xDD:
                return self.decode_index_instruction(addr, 'IX')
            elif opcode == 0xFD:
                return self.decode_index_instruction(addr, 'IY')
            elif opcode in Z80_ONLY_OPCODES:
                return self.decode_z80_only_opcode(addr, opcode)

        if opcode not in OPCODES:
            # Unknown opcode - treat as data
            return ('DB', 1, [format_hex8(opcode)], None)

        mnemonic, size, operand_type = OPCODES[opcode]
        operands = []
        target = None

        if operand_type is None:
            pass
        elif operand_type == ('r', 's'):
            # MOV r,r' (8080) / LD r,r' (Z80)
            dst = REGS[(opcode >> 3) & 7]
            src = REGS[opcode & 7]
            if self.z80_mode:
                mnemonic = 'LD'
                dst = Z80_REGS[(opcode >> 3) & 7]
                src = Z80_REGS[opcode & 7]
            operands = [dst, src]
        elif operand_type == ('r', 'd8'):
            # MVI r,d8 (8080) / LD r,n (Z80)
            reg = REGS[(opcode >> 3) & 7]
            imm = self.byte_at(addr + 1)
            if imm is None:
                return None
            if self.z80_mode:
                mnemonic = 'LD'
                reg = Z80_REGS[(opcode >> 3) & 7]
            operands = [reg, format_hex8(imm)]
        elif operand_type == ('r',):
            # INR/DCR r (8080) / INC/DEC r (Z80)
            reg = REGS[(opcode >> 3) & 7]
            if self.z80_mode:
                reg = Z80_REGS[(opcode >> 3) & 7]
                mnemonic = 'INC' if mnemonic == 'INR' else 'DEC'
            operands = [reg]
        elif operand_type == ('s',):
            # ALU r
            reg = REGS[opcode & 7]
            if self.z80_mode:
                reg = Z80_REGS[opcode & 7]
                # Z80 uses different ALU mnemonics
                mnemonic = self.z80_alu_mnemonic(mnemonic, reg)
            operands = [reg] if not self.z80_mode or mnemonic not in ('ADD', 'ADC', 'SBC') else ['A', reg]
        elif operand_type == ('rp',):
            # Register pair operations
            rp = REGPAIRS[(opcode >> 4) & 3]
            if self.z80_mode:
                rp = Z80_REGPAIRS[(opcode >> 4) & 3]
                if mnemonic == 'INX':
                    mnemonic = 'INC'
                elif mnemonic == 'DCX':
                    mnemonic = 'DEC'
                elif mnemonic == 'DAD':
                    mnemonic = 'ADD'
                    operands = ['HL', rp]
                    return (mnemonic, size, operands, target)
                elif mnemonic == 'LDAX':
                    mnemonic = 'LD'
                    operands = ['A', '(' + rp + ')']
                    return (mnemonic, size, operands, target)
                elif mnemonic == 'STAX':
                    mnemonic = 'LD'
                    operands = ['(' + rp + ')', 'A']
                    return (mnemonic, size, operands, target)
            operands = [rp]
        elif operand_type == ('rp2',):
            # PUSH/POP with PSW
            rp = REGPAIRS_PUSH[(opcode >> 4) & 3]
            if self.z80_mode:
                rp = Z80_REGPAIRS_AF[(opcode >> 4) & 3]
            operands = [rp]
        elif operand_type == ('rp', 'd16'):
            # LXI rp,d16 (8080) / LD rp,nn (Z80)
            rp = REGPAIRS[(opcode >> 4) & 3]
            imm = self.word_at(addr + 1)
            if imm is None:
                return None
            if self.z80_mode:
                mnemonic = 'LD'
                rp = Z80_REGPAIRS[(opcode >> 4) & 3]
            operands = [rp, format_hex16(imm)]
            # Check if this might be a code reference
            if self.org <= imm < self.end:
                target = imm
        elif operand_type == ('d8',):
            # 8-bit immediate
            imm = self.byte_at(addr + 1)
            if imm is None:
                return None
            if self.z80_mode:
                mnemonic, operands = self.z80_imm8_mnemonic(mnemonic, imm)
            else:
                operands = [format_hex8(imm)]
        elif operand_type == ('d16',):
            # 16-bit immediate
            imm = self.word_at(addr + 1)
            if imm is None:
                return None
            operands = [format_hex16(imm)]
        elif operand_type == ('a16',):
            # 16-bit address (jumps/calls/memory)
            target = self.word_at(addr + 1)
            if target is None:
                return None
            if self.z80_mode:
                mnemonic, operands = self.z80_addr_mnemonic(mnemonic, target)
            else:
                operands = [format_hex16(target)]
        elif operand_type == ('rst',):
            # RST vector
            n = (opcode >> 3) & 7
            if self.z80_mode:
                operands = [format_hex8(n * 8)]  # Z80 uses actual address
            else:
                operands = [str(n)]

        # Handle Z80 unconditional no-operand mappings
        if self.z80_mode and operand_type is None:
            mnemonic, operands = self.z80_noarg_mnemonic(mnemonic)

        return (mnemonic, size, operands, target)

    def z80_alu_mnemonic(self, mnemonic, reg):
        """Convert 8080 ALU mnemonic to Z80."""
        mapping = {
            'ADD': 'ADD', 'ADC': 'ADC', 'SUB': 'SUB', 'SBB': 'SBC',
            'ANA': 'AND', 'XRA': 'XOR', 'ORA': 'OR', 'CMP': 'CP'
        }
        return mapping.get(mnemonic, mnemonic)

    def z80_imm8_mnemonic(self, mnemonic, imm):
        """Convert 8080 immediate mnemonic to Z80."""
        mapping = {
            'ADI': ('ADD', ['A', format_hex8(imm)]),
            'ACI': ('ADC', ['A', format_hex8(imm)]),
            'SUI': ('SUB', [format_hex8(imm)]),
            'SBI': ('SBC', ['A', format_hex8(imm)]),
            'ANI': ('AND', [format_hex8(imm)]),
            'XRI': ('XOR', [format_hex8(imm)]),
            'ORI': ('OR', [format_hex8(imm)]),
            'CPI': ('CP', [format_hex8(imm)]),
            'IN': ('IN', ['A', '(' + format_hex8(imm) + ')']),
            'OUT': ('OUT', ['(' + format_hex8(imm) + ')', 'A']),
        }
        return mapping.get(mnemonic, (mnemonic, [format_hex8(imm)]))

    def z80_addr_mnemonic(self, mnemonic, target):
        """Convert 8080 address mnemonic to Z80."""
        addr_str = format_hex16(target)
        mapping = {
            'LDA': ('LD', ['A', '(' + addr_str + ')']),
            'STA': ('LD', ['(' + addr_str + ')', 'A']),
            'LHLD': ('LD', ['HL', '(' + addr_str + ')']),
            'SHLD': ('LD', ['(' + addr_str + ')', 'HL']),
        }
        return mapping.get(mnemonic, (mnemonic, [addr_str]))

    def z80_noarg_mnemonic(self, mnemonic):
        """Convert 8080 no-operand mnemonic to Z80."""
        mapping = {
            'RLC': ('RLCA', []),
            'RRC': ('RRCA', []),
            'RAL': ('RLA', []),
            'RAR': ('RRA', []),
            'CMA': ('CPL', []),
            'STC': ('SCF', []),
            'CMC': ('CCF', []),
            'PCHL': ('JP', ['(HL)']),
            'SPHL': ('LD', ['SP', 'HL']),
            'XCHG': ('EX', ['DE', 'HL']),
            'XTHL': ('EX', ['(SP)', 'HL']),
            'LDAX': ('LD', ['A', '(BC)']),  # handled elsewhere
            'STAX': ('LD', ['(BC)', 'A']),  # handled elsewhere
        }
        return mapping.get(mnemonic, (mnemonic, []))

    def decode_cb_instruction(self, addr):
        """Decode CB-prefixed instruction."""
        opcode2 = self.byte_at(addr + 1)
        if opcode2 is None:
            return None

        if opcode2 not in CB_OPCODES:
            return ('DB', 2, [format_hex8(0xCB), format_hex8(opcode2)], None)

        mnemonic, _, operand_type = CB_OPCODES[opcode2]
        operands = []
        r_idx = opcode2 & 7

        if operand_type == ('cb_r',):
            # Shift/rotate operations
            operands = [Z80_REGS[r_idx]]
        elif operand_type == ('bit_n', 'cb_r'):
            # BIT/RES/SET
            n = (opcode2 >> 3) & 7
            operands = [str(n), Z80_REGS[r_idx]]

        return (mnemonic, 2, operands, None)

    def decode_ed_instruction(self, addr):
        """Decode ED-prefixed instruction."""
        opcode2 = self.byte_at(addr + 1)
        if opcode2 is None:
            return None

        if opcode2 not in ED_OPCODES:
            # Unknown ED opcode - treat as NOP NOP
            return ('DB', 2, [format_hex8(0xED), format_hex8(opcode2)], None)

        mnemonic, size, operand_type = ED_OPCODES[opcode2]
        operands = []
        target = None

        if operand_type is None:
            pass
        elif operand_type == ('ed_r', '(C)'):
            r_idx = (opcode2 >> 3) & 7
            operands = [Z80_REGS[r_idx], '(C)']
        elif operand_type == ('(C)', 'ed_r'):
            r_idx = (opcode2 >> 3) & 7
            operands = ['(C)', Z80_REGS[r_idx]]
        elif operand_type == ('(C)',):
            operands = ['(C)']
        elif operand_type == ('HL', 'ed_rp'):
            rp_idx = (opcode2 >> 4) & 3
            operands = ['HL', Z80_REGPAIRS[rp_idx]]
        elif operand_type == ('(a16)', 'ed_rp'):
            rp_idx = (opcode2 >> 4) & 3
            target = self.word_at(addr + 2)
            if target is None:
                return None
            operands = ['(' + format_hex16(target) + ')', Z80_REGPAIRS[rp_idx]]
        elif operand_type == ('ed_rp', '(a16)'):
            rp_idx = (opcode2 >> 4) & 3
            target = self.word_at(addr + 2)
            if target is None:
                return None
            operands = [Z80_REGPAIRS[rp_idx], '(' + format_hex16(target) + ')']
        elif operand_type == ('im0',):
            operands = ['0']
        elif operand_type == ('im1',):
            operands = ['1']
        elif operand_type == ('im2',):
            operands = ['2']
        elif len(operand_type) == 2 and isinstance(operand_type[0], str) and isinstance(operand_type[1], str):
            # Simple register pair like ('I', 'A')
            operands = list(operand_type)

        return (mnemonic, size, operands, target)

    def decode_index_instruction(self, addr, index_reg):
        """Decode DD/FD-prefixed instruction (IX/IY)."""
        opcode2 = self.byte_at(addr + 1)
        if opcode2 is None:
            return None

        # DD CB / FD CB - indexed bit operations
        if opcode2 == 0xCB:
            return self.decode_indexed_cb_instruction(addr, index_reg)

        # Map H->IXH/IYH, L->IXL/IYL, (HL)->(IX+d)/(IY+d)
        ih = index_reg + 'H'
        il = index_reg + 'L'

        # Check for instructions that use index register
        if opcode2 in OPCODES:
            mnemonic, size, operand_type = OPCODES[opcode2]

            # Instructions that use (HL) get displacement
            uses_hl_indirect = False
            if operand_type in [('r', 's'), ('r', 'd8'), ('r',), ('s',)]:
                r_bits = (opcode2 >> 3) & 7 if operand_type != ('s',) else opcode2 & 7
                s_bits = opcode2 & 7
                if r_bits == 6 or (operand_type in [('r', 's'), ('s',)] and s_bits == 6):
                    uses_hl_indirect = True

            if uses_hl_indirect:
                # These need displacement byte
                disp = self.byte_at(addr + 2)
                if disp is None:
                    return None
                disp_signed = disp if disp < 128 else disp - 256
                idx_operand = f'({index_reg}{disp_signed:+d})'

                if operand_type == ('r', 's'):
                    # LD r,(IX+d) or LD (IX+d),r
                    d = (opcode2 >> 3) & 7
                    s = opcode2 & 7
                    if d == 6:  # LD (IX+d),r
                        dst = idx_operand
                        src = Z80_REGS[s]
                    else:  # LD r,(IX+d)
                        dst = Z80_REGS[d]
                        src = idx_operand
                    return ('LD', 3, [dst, src], None)
                elif operand_type == ('r', 'd8'):
                    # LD (IX+d),n
                    imm = self.byte_at(addr + 3)
                    if imm is None:
                        return None
                    return ('LD', 4, [idx_operand, format_hex8(imm)], None)
                elif operand_type == ('r',):
                    # INC/DEC (IX+d)
                    mnemonic = 'INC' if mnemonic == 'INR' else 'DEC'
                    return (mnemonic, 3, [idx_operand], None)
                elif operand_type == ('s',):
                    # ALU (IX+d)
                    z80_mn = self.z80_alu_mnemonic(mnemonic, idx_operand)
                    if z80_mn in ('ADD', 'ADC', 'SBC'):
                        return (z80_mn, 3, ['A', idx_operand], None)
                    return (z80_mn, 3, [idx_operand], None)

            # Instructions using HL register pair
            if operand_type == ('rp',) or operand_type == ('rp2',):
                rp_idx = (opcode2 >> 4) & 3
                if rp_idx == 2:  # HL -> IX/IY
                    rp = index_reg
                else:
                    rp = Z80_REGPAIRS[rp_idx] if operand_type == ('rp',) else Z80_REGPAIRS_AF[rp_idx]

                if mnemonic == 'INX':
                    return ('INC', 2, [rp], None)
                elif mnemonic == 'DCX':
                    return ('DEC', 2, [rp], None)
                elif mnemonic == 'DAD':
                    src_rp = Z80_REGPAIRS[rp_idx]
                    if rp_idx == 2:
                        src_rp = index_reg
                    return ('ADD', 2, [index_reg, src_rp], None)
                elif mnemonic in ('PUSH', 'POP'):
                    return (mnemonic, 2, [rp], None)

            elif operand_type == ('rp', 'd16'):
                rp_idx = (opcode2 >> 4) & 3
                if rp_idx == 2:  # HL -> IX/IY
                    imm = self.word_at(addr + 2)
                    if imm is None:
                        return None
                    return ('LD', 4, [index_reg, format_hex16(imm)], None)

            # Undocumented: H/L -> IXH/IXL/IYH/IYL
            if operand_type == ('r', 's'):
                d = (opcode2 >> 3) & 7
                s = opcode2 & 7
                # Undocumented LD with IXH/IXL/IYH/IYL
                if d in (4, 5) or s in (4, 5):
                    dst_map = {4: ih, 5: il}
                    src_map = {4: ih, 5: il}
                    dst = dst_map.get(d, Z80_REGS[d])
                    src = src_map.get(s, Z80_REGS[s])
                    return ('LD', 2, [dst, src], None)

            if operand_type == ('r', 'd8'):
                r = (opcode2 >> 3) & 7
                if r in (4, 5):
                    imm = self.byte_at(addr + 2)
                    if imm is None:
                        return None
                    reg = ih if r == 4 else il
                    return ('LD', 3, [reg, format_hex8(imm)], None)

            if operand_type == ('r',):
                r = (opcode2 >> 3) & 7
                if r in (4, 5):
                    reg = ih if r == 4 else il
                    mnemonic = 'INC' if mnemonic == 'INR' else 'DEC'
                    return (mnemonic, 2, [reg], None)

            if operand_type == ('s',):
                r = opcode2 & 7
                if r in (4, 5):
                    reg = ih if r == 4 else il
                    z80_mn = self.z80_alu_mnemonic(mnemonic, reg)
                    if z80_mn in ('ADD', 'ADC', 'SBC'):
                        return (z80_mn, 2, ['A', reg], None)
                    return (z80_mn, 2, [reg], None)

        # Special cases
        if opcode2 == 0xE9:  # JP (IX)/(IY)
            return ('JP', 2, ['(' + index_reg + ')'], None)
        if opcode2 == 0xF9:  # LD SP,IX/IY
            return ('LD', 2, ['SP', index_reg], None)
        if opcode2 == 0xE3:  # EX (SP),IX/IY
            return ('EX', 2, ['(SP)', index_reg], None)
        if opcode2 == 0x2A:  # LD IX/IY,(nn)
            target = self.word_at(addr + 2)
            if target is None:
                return None
            return ('LD', 4, [index_reg, '(' + format_hex16(target) + ')'], target)
        if opcode2 == 0x22:  # LD (nn),IX/IY
            target = self.word_at(addr + 2)
            if target is None:
                return None
            return ('LD', 4, ['(' + format_hex16(target) + ')', index_reg], target)

        # Prefix has no effect on this opcode - output as prefix + instruction
        # (undocumented behavior)
        return ('DB', 1, [format_hex8(0xDD if index_reg == 'IX' else 0xFD)], None)

    def decode_indexed_cb_instruction(self, addr, index_reg):
        """Decode DD CB / FD CB indexed bit operations."""
        disp = self.byte_at(addr + 2)
        opcode3 = self.byte_at(addr + 3)
        if disp is None or opcode3 is None:
            return None

        disp_signed = disp if disp < 128 else disp - 256
        idx_operand = f'({index_reg}{disp_signed:+d})'
        r_idx = opcode3 & 7

        # Determine operation
        if opcode3 < 0x40:
            # Rotate/shift
            op_idx = (opcode3 >> 3) & 7
            mnemonic = CB_OPS[op_idx]
            if r_idx == 6:
                return (mnemonic, 4, [idx_operand], None)
            else:
                # Undocumented: result also stored in register
                return (mnemonic, 4, [idx_operand, Z80_REGS[r_idx]], None)
        elif opcode3 < 0x80:
            # BIT
            n = (opcode3 >> 3) & 7
            return ('BIT', 4, [str(n), idx_operand], None)
        elif opcode3 < 0xC0:
            # RES
            n = (opcode3 >> 3) & 7
            if r_idx == 6:
                return ('RES', 4, [str(n), idx_operand], None)
            else:
                # Undocumented: result also stored in register
                return ('RES', 4, [str(n), idx_operand, Z80_REGS[r_idx]], None)
        else:
            # SET
            n = (opcode3 >> 3) & 7
            if r_idx == 6:
                return ('SET', 4, [str(n), idx_operand], None)
            else:
                # Undocumented: result also stored in register
                return ('SET', 4, [str(n), idx_operand, Z80_REGS[r_idx]], None)

    def decode_z80_only_opcode(self, addr, opcode):
        """Decode Z80-only base opcodes (JR, DJNZ, EX AF,AF')."""
        mnemonic, size, operand_type = Z80_ONLY_OPCODES[opcode]
        operands = []
        target = None

        if operand_type == ('af_af',):
            # EX AF,AF'
            operands = ['AF', "AF'"]
        elif operand_type == ('rel8',):
            # JR d or DJNZ d
            disp = self.byte_at(addr + 1)
            if disp is None:
                return None
            disp_signed = disp if disp < 128 else disp - 256
            target = addr + 2 + disp_signed
            operands = [format_hex16(target)]
        elif len(operand_type) == 2 and operand_type[1] == 'rel8':
            # JR cc,d
            cond = operand_type[0]
            disp = self.byte_at(addr + 1)
            if disp is None:
                return None
            disp_signed = disp if disp < 128 else disp - 256
            target = addr + 2 + disp_signed
            operands = [cond, format_hex16(target)]

        return (mnemonic, size, operands, target)

    def is_unconditional_transfer(self, mnemonic):
        """Check if instruction is unconditional transfer of control."""
        # Include both 8080 and Z80 mnemonics
        return mnemonic in ('JMP', 'RET', 'PCHL', 'HLT', 'JP', 'RETI', 'RETN')

    def is_call(self, mnemonic):
        """Check if instruction is a call."""
        return mnemonic in ('CALL', 'CNZ', 'CZ', 'CNC', 'CC', 'CPO', 'CPE', 'CP', 'CM', 'RST')

    def is_jump(self, mnemonic):
        """Check if instruction is a jump."""
        return mnemonic in ('JMP', 'JNZ', 'JZ', 'JNC', 'JC', 'JPO', 'JPE', 'JP', 'JM', 'PCHL')

    def is_in_post_jump_gap(self, addr):
        """Check if address falls within a post-jump data gap."""
        for gap_start, gap_end in self.post_jump_gaps:
            if gap_start <= addr < gap_end:
                return True
        return False

    def analyze_code_flow(self, entry_points):
        """Trace code flow from entry points to identify code regions.

        Uses a two-pass approach:
        1. First pass from primary entry points, recording post-jump gaps
        2. Second pass processes discovered targets, skipping those in gaps
        """
        # First pass: trace from primary entry points and record post-jump gaps
        primary_targets = set()  # Targets discovered in first pass
        work_list = list(entry_points)
        visited = set()

        while work_list:
            addr = work_list.pop()

            if addr in visited:
                continue
            if addr < self.org or addr >= self.end:
                continue
            if addr in self.data_addrs:
                continue  # Don't trace into forced data regions
            if addr in self.table_addrs:
                continue  # Don't trace into address tables

            visited.add(addr)

            while addr < self.end:
                if addr in self.code_addrs:
                    break
                if addr in self.data_addrs:
                    break  # Stop at forced data region
                if addr in self.table_addrs:
                    break  # Stop at address table

                result = self.decode_instruction(addr)
                if result is None:
                    break

                mnemonic, size, operands, target = result

                # Mark as code
                for i in range(size):
                    self.code_addrs.add(addr + i)

                # Track references
                if target is not None and self.org <= target < self.end:
                    self.refs_from[target].add(addr)
                    self.refs_to[addr].add(target)

                    if self.is_jump(mnemonic) or self.is_call(mnemonic):
                        if target not in visited:
                            primary_targets.add(target)
                            work_list.append(target)

                # After unconditional transfer, record the gap
                if self.is_unconditional_transfer(mnemonic):
                    gap_start = addr + size
                    # Find next known code or end of file
                    gap_end = self.end
                    # The gap extends until we hit known code or a branch target
                    self.post_jump_gaps.append((gap_start, gap_end))
                    break

                addr += size

        # Refine post-jump gaps: end each gap at the next known branch target
        refined_gaps = []
        for gap_start, gap_end in self.post_jump_gaps:
            # Find the lowest target address that's >= gap_start
            next_target = gap_end
            for target in primary_targets:
                if gap_start <= target < next_target:
                    next_target = target
            # Also check entry points
            for ep in entry_points:
                if gap_start <= ep < next_target:
                    next_target = ep
            if gap_start < next_target:
                refined_gaps.append((gap_start, next_target))
        self.post_jump_gaps = refined_gaps

        # Mark post-jump gap addresses as data (unless already code)
        for gap_start, gap_end in self.post_jump_gaps:
            for addr in range(gap_start, gap_end):
                if addr not in self.code_addrs:
                    self.data_addrs.add(addr)

    def find_strings(self, min_len=4):
        """Find potential ASCII strings in non-code areas."""
        i = 0
        while i < len(self.data):
            addr = self.org + i
            if addr in self.code_addrs:
                i += 1
                continue

            # Look for printable ASCII sequence
            start = i
            while i < len(self.data):
                b = self.data[i]
                # Accept printable ASCII and common control chars
                if 0x20 <= b <= 0x7E or b in (0x0D, 0x0A, 0x09):
                    i += 1
                elif b == 0 and i > start:  # Null terminator
                    i += 1
                    break
                elif b & 0x80 and 0x20 <= (b & 0x7F) <= 0x7E:
                    # High bit set - common in BASIC for tokenized text
                    i += 1
                    break
                else:
                    break

            if i - start >= min_len:
                self.strings[self.org + start] = self.data[start:i]
            elif i == start:
                i += 1

    def generate_labels(self):
        """Generate labels for all referenced addresses."""
        # First, apply user-specified labels
        for addr, name in self.user_labels.items():
            self.labels[addr] = name

        # Build set of instruction start addresses
        instr_starts = set()
        addr = self.org
        while addr < self.end:
            if addr in self.code_addrs:
                instr_starts.add(addr)
                result = self.decode_instruction(addr)
                if result:
                    addr += result[1]  # Skip instruction size
                else:
                    addr += 1
            else:
                addr += 1

        # Entry points (don't override user labels)
        for addr in self.entry_points:
            if addr not in self.labels and addr in instr_starts:
                self.labels[addr] = f'L{addr:04X}'

        # Jump/call targets - only create labels at instruction boundaries
        for addr in self.refs_from:
            if addr not in self.labels and self.org <= addr < self.end:
                # Only create label if it's at an instruction start or data location
                if addr in instr_starts:
                    self.labels[addr] = f'L{addr:04X}'
                elif addr not in self.code_addrs:
                    # Data reference
                    self.labels[addr] = f'D{addr:04X}'
                # If it's mid-instruction, don't create a label

    def format_operand(self, op, addr):
        """Format operand, using label if available."""
        if op.endswith('H'):
            hex_part = op[:-1]
            # Check length before stripping leading 0
            if len(hex_part) == 4:
                # 16-bit address - strip leading 0 for parsing if present
                if hex_part.startswith('0'):
                    hex_part = hex_part[1:]
                target = int(hex_part, 16)
                if target in self.labels:
                    return self.labels[target]
                # If target is within our range but has no label, keep as hex
                # (it won't cause undefined symbol errors in assembler)
        return op

    def is_printable(self, b):
        """Check if byte is printable ASCII (0x20-0x7E)."""
        return 0x20 <= b <= 0x7E

    def format_dc_string(self, start_addr):
        """Format a high-bit terminated string starting at addr.

        Returns (formatted_string, bytes_consumed, end_char_with_high_bit).
        The string ends when we encounter a byte with bit 7 set where
        the lower 7 bits are printable.
        """
        chars = []
        addr = start_addr
        while addr < self.end and addr in self.dc_string_addrs:
            b = self.byte_at(addr)
            if b is None:
                break

            if b & 0x80:
                # High bit set - check if lower 7 bits are printable
                low = b & 0x7F
                if self.is_printable(low):
                    # This is the terminator
                    if chars:
                        return (chars, addr - start_addr, chr(low))
                    else:
                        # Single char with high bit
                        return ([], 0, chr(low))
                else:
                    # High bit set but not printable - end the scan
                    break
            elif self.is_printable(b):
                chars.append(chr(b))
                addr += 1
            else:
                # Non-printable without high bit - end the scan
                break

        # No terminator found - return what we have (if any)
        if chars:
            return (chars, addr - start_addr, None)
        return ([], 0, None)

    def format_da_string(self, start_addr):
        """Format an ASCII string starting at addr.

        Handles both null-terminated and fixed-length strings based on da_string_len.
        Returns (string_content, bytes_consumed).
        """
        fixed_len = self.da_string_len.get(start_addr, 0)

        if fixed_len > 0:
            # Fixed-length string
            chars = []
            bytes_consumed = 0
            for i in range(fixed_len):
                addr = start_addr + i
                if addr >= self.end or addr not in self.da_string_addrs:
                    break
                b = self.byte_at(addr)
                if b is None:
                    break
                bytes_consumed += 1
                if self.is_printable(b):
                    chars.append(chr(b))
                else:
                    # Non-printable in fixed string - end string here
                    break
            return (''.join(chars), bytes_consumed)
        else:
            # Null-terminated string
            chars = []
            addr = start_addr
            while addr < self.end and addr in self.da_string_addrs:
                b = self.byte_at(addr)
                if b is None:
                    break

                if b == 0:
                    # Null terminator
                    return (''.join(chars), addr - start_addr + 1)
                elif self.is_printable(b):
                    chars.append(chr(b))
                    addr += 1
                else:
                    # Non-printable - end the scan
                    break

            # No null found - return what we have
            if chars:
                return (''.join(chars), addr - start_addr)
            return ('', 0)

    def detect_dispatch_tables(self):
        """Detect jump/dispatch tables by looking for the dispatch pattern in code.

        Pattern: LXI H,addr followed by DAD B (or DAD D) indicates addr is a table base.
        The pattern 'lxi h,TABLE / dad rp' is used to calculate table[offset].
        """
        # Scan code for LXI H,addr followed by DAD B/D pattern
        for addr in sorted(self.code_addrs):
            opcode = self.byte_at(addr)
            if opcode == 0x21:  # LXI H,d16
                table_addr = self.word_at(addr + 1)
                if table_addr is None or table_addr < self.org or table_addr >= self.end:
                    continue

                # Check if next instruction is DAD B (0x09) or DAD D (0x19)
                next_addr = addr + 3
                next_opcode = self.byte_at(next_addr)
                if next_opcode in (0x09, 0x19):  # DAD B or DAD D
                    # This looks like a table dispatch pattern
                    # Now verify it's actually a table by checking if entries are valid addresses
                    if self.verify_address_table(table_addr):
                        self.mark_address_table(table_addr)

    def detect_post_jump_tables(self):
        """Detect address tables immediately following unconditional jumps at entry points.

        Common pattern: JMP main_code followed by dispatch table addresses.
        """
        for ep in self.entry_points:
            result = self.decode_instruction(ep)
            if result is None:
                continue

            mnemonic, size, operands, target = result
            if mnemonic == 'JMP' and target is not None:
                # Check if bytes after JMP form an address table
                table_start = ep + size
                if self.verify_address_table(table_start, min_entries=3):
                    self.mark_address_table(table_start)

    def verify_address_table(self, start_addr, min_entries=3):
        """Verify that a sequence of bytes looks like an address table.

        Returns True if we find at least min_entries consecutive valid addresses
        that point within our code range.
        """
        valid_count = 0
        addr = start_addr

        while addr + 1 < self.end:
            word = self.word_at(addr)
            if word is None:
                break

            # Check if this looks like a valid code/data address
            if self.org <= word < self.end:
                # Additional heuristic: the target should be at an instruction boundary
                # or in a data area (not mid-instruction)
                valid_count += 1
                addr += 2
            else:
                # Not a valid address - end of table or not a table
                break

            # Don't scan forever - tables are typically < 256 entries
            if valid_count > 256:
                break

        return valid_count >= min_entries

    def find_table_end(self, start_addr):
        """Find the end of an address table starting at start_addr.

        Returns the address after the last valid table entry.
        """
        addr = start_addr
        last_valid = start_addr
        consecutive_invalid = 0

        while addr + 1 < self.end and consecutive_invalid < 2:
            word = self.word_at(addr)
            if word is None:
                break

            if self.org <= word < self.end:
                last_valid = addr + 2
                consecutive_invalid = 0
            else:
                consecutive_invalid += 1

            addr += 2

            # Safety limit
            if addr - start_addr > 512:
                break

        return last_valid

    def mark_address_table(self, start_addr):
        """Mark a range as an address table."""
        end_addr = self.find_table_end(start_addr)

        for addr in range(start_addr, end_addr):
            self.detected_tables.add(addr)
            self.table_addrs.add(addr)
            self.data_addrs.add(addr)

    def scan_address_tables(self):
        """Scan address tables and add their contents to refs_from for label generation.

        Returns a set of addresses found in tables that should be traced as code.
        """
        table_targets = set()

        # Combine user-specified and auto-detected tables
        all_tables = self.table_addrs | self.detected_tables

        # Get sorted list of table addresses
        table_list = sorted(all_tables)
        i = 0
        while i < len(table_list) - 1:
            addr = table_list[i]
            # Check if next byte is also in table (we need pairs for DW)
            if i + 1 < len(table_list) and table_list[i + 1] == addr + 1:
                word = self.word_at(addr)
                if word is not None and self.org <= word < self.end:
                    # This looks like an address - add to refs_from
                    self.refs_from[word].add(addr)
                    # Collect as potential code target
                    table_targets.add(word)
                i += 2
            else:
                i += 1

        return table_targets

    def disassemble(self, entry_points=None):
        """Perform full disassembly."""
        # Default entry point is start of COM file
        if entry_points is None:
            entry_points = [self.org]
        self.entry_points = set(entry_points)

        # First, detect tables after entry point JMPs (before code flow analysis)
        self.detect_post_jump_tables()

        # Analyze code flow from initial entry points
        self.analyze_code_flow(entry_points)

        # Detect dispatch tables by finding LXI H,addr / DAD pattern in code
        self.detect_dispatch_tables()

        # Scan address tables and get their targets
        table_targets = self.scan_address_tables()

        # Trace table targets as code entry points (they're jump destinations)
        # Filter to those not already traced
        new_entries = table_targets - self.code_addrs
        if new_entries:
            # Table targets take precedence - remove from data_addrs if present
            # (they might have been marked as data by post-jump gap detection)
            self.data_addrs -= new_entries
            self.entry_points.update(new_entries)
            self.analyze_code_flow(list(new_entries))

            # Check for more dispatch tables in newly discovered code
            self.detect_dispatch_tables()

            # Scan any new tables found
            more_targets = self.scan_address_tables()
            more_new = more_targets - self.code_addrs
            if more_new:
                self.data_addrs -= more_new
                self.entry_points.update(more_new)
                self.analyze_code_flow(list(more_new))

        # Generate labels
        self.generate_labels()

        # Find strings in data areas
        self.find_strings()

        # Generate output
        return self.generate_output()

    def generate_output(self):
        """Generate .MAC output."""
        lines = []
        lines.append('; Disassembled by ud80')
        lines.append(f'; Source file size: {len(self.data)} bytes')
        lines.append(f'; Mode: {"Z80" if self.z80_mode else "8080"}')
        lines.append('')
        lines.append('\t.Z80' if self.z80_mode else '\t.8080')
        lines.append('')
        lines.append(f'\tORG\t{format_hex16(self.org)}')
        lines.append('')

        addr = self.org
        while addr < self.end:
            line_parts = []

            # Add label if present
            if addr in self.labels:
                line_parts.append(f'{self.labels[addr]}:')
            else:
                line_parts.append('')

            if addr in self.code_addrs:
                # Disassemble instruction
                result = self.decode_instruction(addr)
                if result:
                    mnemonic, size, operands, target = result

                    # Format operands with labels
                    formatted_ops = [self.format_operand(op, addr) for op in operands]

                    if formatted_ops:
                        line_parts.append(f'\t{mnemonic}\t{",".join(formatted_ops)}')
                    else:
                        line_parts.append(f'\t{mnemonic}')

                    # Add address comment
                    bytes_str = ' '.join(f'{self.byte_at(addr+i):02X}' for i in range(size))
                    line_parts.append(f'\t; {addr:04X}: {bytes_str}')

                    lines.append(''.join(line_parts))
                    addr += size
                else:
                    # Shouldn't happen, but handle it
                    b = self.byte_at(addr)
                    line_parts.append(f'\tDB\t{format_hex8(b)}')
                    line_parts.append(f'\t; {addr:04X}')
                    lines.append(''.join(line_parts))
                    addr += 1
            elif addr in self.table_addrs and addr + 1 in self.table_addrs:
                # Address table - output as DW with label if possible
                word = self.word_at(addr)
                if word is not None:
                    if word in self.labels:
                        line_parts.append(f'\tDW\t{self.labels[word]}')
                    else:
                        line_parts.append(f'\tDW\t{format_hex16(word)}')
                    line_parts.append(f'\t; {addr:04X}: {self.byte_at(addr):02X} {self.byte_at(addr+1):02X}')
                    lines.append(''.join(line_parts))
                    addr += 2
                else:
                    # Shouldn't happen, but handle edge case
                    b = self.byte_at(addr)
                    line_parts.append(f'\tDB\t{format_hex8(b)}')
                    line_parts.append(f'\t; {addr:04X}')
                    lines.append(''.join(line_parts))
                    addr += 1
            elif addr in self.dc_string_addrs:
                # High-bit terminated string - use DC directive
                chars, consumed, end_char = self.format_dc_string(addr)
                if chars or end_char:
                    # Build the complete string (chars + end_char)
                    full_string = ''.join(chars)
                    if end_char:
                        full_string += end_char
                    # Escape single quotes
                    s = full_string.replace("'", "''")
                    # Use DC directive which sets high bit on last char
                    line_parts.append(f"\tDC\t'{s}'")
                    bytes_used = consumed + (1 if end_char else 0)
                    bytes_str = ' '.join(f'{self.byte_at(addr+i):02X}' for i in range(bytes_used))
                    line_parts.append(f'\t; {addr:04X}: {bytes_str}')
                    lines.append(''.join(line_parts))
                    addr += bytes_used
                else:
                    # Can't parse as string - output as byte
                    b = self.byte_at(addr)
                    line_parts.append(f'\tDB\t{format_hex8(b)}')
                    line_parts.append(f'\t; {addr:04X}')
                    lines.append(''.join(line_parts))
                    addr += 1
            elif addr in self.da_string_addrs:
                # ASCII string (null-terminated or fixed-length)
                content, consumed = self.format_da_string(addr)
                if content:
                    # Escape single quotes
                    s = content.replace("'", "''")
                    # Output just the string, any null terminator is output separately
                    line_parts.append(f"\tDB\t'{s}'")
                    # Only show string bytes in comment (not the null if present)
                    str_len = len(content)
                    bytes_str = ' '.join(f'{self.byte_at(addr+i):02X}' for i in range(str_len))
                    line_parts.append(f'\t; {addr:04X}: {bytes_str}')
                    lines.append(''.join(line_parts))
                    addr += str_len
                else:
                    # Can't parse as string - output as byte
                    b = self.byte_at(addr)
                    line_parts.append(f'\tDB\t{format_hex8(b)}')
                    line_parts.append(f'\t; {addr:04X}')
                    lines.append(''.join(line_parts))
                    addr += 1
            else:
                # Data byte - output one at a time to allow labels at any position
                b = self.byte_at(addr)
                line_parts.append(f'\tDB\t{format_hex8(b)}')
                line_parts.append(f'\t; {addr:04X}')
                lines.append(''.join(line_parts))
                addr += 1

        lines.append('')
        lines.append('\tEND')

        return '\n'.join(lines)


def parse_range(s):
    """Parse address range like '1000-2000' or '1000'."""
    if '-' in s:
        start, end = s.split('-', 1)
        return (int(start, 16), int(end, 16))
    else:
        return (int(s, 16), int(s, 16))


def main():
    parser = argparse.ArgumentParser(description='ud80 - 8080/Z80 Disassembler for CP/M .COM files')
    parser.add_argument('-v', '--version', action='version', version=f'%(prog)s {__version__}')
    parser.add_argument('input', help='Input .COM file')
    parser.add_argument('-o', '--output', help='Output .MAC file')
    parser.add_argument('-e', '--entry', action='append', metavar='ADDR[,LABEL]',
                       help='Additional entry point. Format: hex address, optionally with label name '
                            '(e.g., -e 2024 or -e 2024,myloop)')
    parser.add_argument('-l', '--label', action='append', metavar='ADDR[,LABEL]',
                       help='Add a label at address without tracing as code. Format: hex address, '
                            'optionally with label name (e.g., -l 1A00 or -l 1A00,buffer)')
    parser.add_argument('-d', '--data', action='append', metavar='RANGE',
                       help='Force address range to be data, not code (e.g., -d 1A00-1A7F). '
                            'Can be repeated for multiple ranges.')
    parser.add_argument('-t', '--table', action='append', metavar='RANGE',
                       help='Address table range, output as DW with labels (e.g., -t 0103-0120). '
                            'Can be repeated for multiple ranges.')
    parser.add_argument('-dc', '--dcstring', action='append', metavar='RANGE',
                       help='High-bit terminated string table (e.g., -dc 0200-02FF). '
                            'Strings end with last char OR 80H. Can be repeated.')
    parser.add_argument('-da', '--dastring', action='append', metavar='RANGE[,LEN]',
                       help='ASCII string region. Format: RANGE or RANGE,LEN. '
                            'Without LEN: strings end at null (00H) or non-printable. '
                            'With LEN: fixed-length strings. Can be repeated.')
    parser.add_argument('--org', default='0100',
                       help='Origin address (default: 0100 for CP/M)')
    parser.add_argument('-z', '--z80', action='store_true',
                       help='Z80 mode: use Zilog mnemonics and decode Z80-only instructions')

    args = parser.parse_args()

    # Read input file
    with open(args.input, 'rb') as f:
        data = f.read()

    org = int(args.org, 16)

    # Create disassembler
    disasm = Disassembler(data, org, z80_mode=args.z80)

    # Parse entry points (format: addr or addr,label)
    entry_points = [org]
    if args.entry:
        for e in args.entry:
            if ',' in e:
                addr_str, label = e.split(',', 1)
                addr = int(addr_str, 16)
                disasm.user_labels[addr] = label
            else:
                addr = int(e, 16)
            entry_points.append(addr)

    # Parse label-only points (format: addr or addr,label) - no code tracing
    if args.label:
        for l in args.label:
            if ',' in l:
                addr_str, label = l.split(',', 1)
                addr = int(addr_str, 16)
                disasm.user_labels[addr] = label
            else:
                addr = int(l, 16)
                # Auto-generate label if not specified
                disasm.user_labels[addr] = f'L{addr:04X}'

    # Parse data ranges
    if args.data:
        for d in args.data:
            start, end = parse_range(d)
            for addr in range(start, end + 1):
                disasm.data_addrs.add(addr)

    # Parse table ranges (also mark as data to prevent code tracing)
    if args.table:
        for t in args.table:
            start, end = parse_range(t)
            for addr in range(start, end + 1):
                disasm.table_addrs.add(addr)
                disasm.data_addrs.add(addr)

    # Parse DC string ranges (high-bit terminated strings)
    if args.dcstring:
        for dc in args.dcstring:
            start, end = parse_range(dc)
            for addr in range(start, end + 1):
                disasm.dc_string_addrs.add(addr)
                disasm.data_addrs.add(addr)

    # Parse DA string ranges (null-terminated or fixed-length ASCII strings)
    if args.dastring:
        for da in args.dastring:
            # Check for optional length: RANGE,LEN
            if ',' in da:
                range_part, len_part = da.rsplit(',', 1)
                try:
                    fixed_len = int(len_part)
                except ValueError:
                    print(f'Error: Invalid length in -da {da}', file=sys.stderr)
                    sys.exit(1)
            else:
                range_part = da
                fixed_len = 0  # null-terminated
            start, end = parse_range(range_part)
            for addr in range(start, end + 1):
                disasm.da_string_addrs.add(addr)
                disasm.da_string_len[addr] = fixed_len
                disasm.data_addrs.add(addr)

    # Disassemble
    output = disasm.disassemble(entry_points)

    # Write output
    if args.output:
        output_path = args.output
    else:
        base = os.path.splitext(args.input)[0]
        output_path = base + '.mac'

    with open(output_path, 'w') as f:
        f.write(output)

    mode = 'Z80' if args.z80 else '8080'
    print(f'Disassembled {args.input} -> {output_path} ({mode} mode)')
    print(f'  Code bytes: {len(disasm.code_addrs)}')
    print(f'  Data bytes: {len(data) - len(disasm.code_addrs)}')
    print(f'  Labels: {len(disasm.labels)}')
    if disasm.detected_tables:
        print(f'  Auto-detected table bytes: {len(disasm.detected_tables)}')


if __name__ == '__main__':
    main()
